import os
from typing import Dict

from pddl_plus_parser.lisp_parsers import DomainParser, ProblemParser, TrajectoryParser
from pddl_plus_parser.models import Domain, Problem, Observation, PDDLObject, MultiAgentObservation
from pytest import fixture

from sam_learning.learners import SAMLearner
from tests.consts import (
    ELEVATORS_DOMAIN_PATH,
    ELEVATORS_PROBLEM_PATH,
    ELEVATORS_TRAJECTORY_PATH,
    WOODWORKING_PROBLEM_PATH,
    WOODWORKING_TRAJECTORY_PATH,
    WOODWORKING_COMBINED_DOMAIN_PATH,
    WOODWORKING_COMBINED_PROBLEM_PATH,
    WOODWORKING_COMBINED_TRAJECTORY_PATH,
    ROVERS_COMBINED_DOMAIN_PATH,
    ROVERS_COMBINED_PROBLEM_PATH,
    ROVERS_COMBINED_TRAJECTORY_PATH,
    LOGISTICS_DOMAIN_PATH,
    SPIDER_DOMAIN_PATH,
    SPIDER_PROBLEM_PATH,
    SPIDER_TRAJECTORY_PATH,
    DEPOTS_NUMERIC_DOMAIN_PATH,
    DEPOTS_NUMERIC_PROBLEM_PATH,
    DEPOT_NUMERIC_TRAJECTORY_PATH,
    NURIKABE_DOMAIN_PATH,
    MINECRAFT_DOMAIN_PATH,
    MINECRAFT_PROBLEM_PATH,
    MINECRAFT_TRAJECTORY_PATH,
    SATELLITE_DOMAIN_PATH,
    SATELLITE_PROBLEM_PATH,
    SATELLITE_NUMERIC_TRAJECTORY_PATH,
    NURIKABE_PROBLEM_PATH,
    NURIKABE_TRAJECTORY_PATH,
    ADL_SATELLITE_DOMAIN_PATH,
    ADL_SATELLITE_PROBLEM_PATH,
    ADL_SATELLITE_TRAJECTORY_PATH,
    DEPOTS_DISCRETE_DOMAIN_PATH,
    DEPOTS_DISCRETE_PROBLEM_PATH,
    MINECRAFT_LARGE_DOMAIN_PATH,
    MINECRAFT_LARGE_PROBLEM_PATH,
    MINECRAFT_LARGE_TRAJECTORY_PATH,
    DRIVERLOG_COMBINED_DOMAIN_PATH,
    DRIVERLOG_COMBINED_PROBLEM_PATH,
    DRIVERLOG_COMBINED_TRAJECTORY_PATH,
    ROVERS_COMBINED_WITH_MACRO_DOMAIN_PATH,
    WOODWORKING_DOMAIN_PATH,
)
from utilities import NegativePreconditionPolicy

os.environ["CONVEX_HULL_ERROR_PATH"] = "tests\\convex_hull_error.txt"
DRIVERLOG_AGENT_NAMES = [f"driver{i}" for i in range(1, 9)]


@fixture()
def elevators_domain() -> Domain:
    domain_parser = DomainParser(ELEVATORS_DOMAIN_PATH, partial_parsing=True)
    return domain_parser.parse_domain()


@fixture()
def elevators_problem(elevators_domain: Domain) -> Problem:
    return ProblemParser(problem_path=ELEVATORS_PROBLEM_PATH, domain=elevators_domain).parse_problem()


@fixture()
def elevators_observation(elevators_domain: Domain, elevators_problem: Problem) -> Observation:
    return TrajectoryParser(elevators_domain, elevators_problem).parse_trajectory(ELEVATORS_TRAJECTORY_PATH)


@fixture()
def elevators_sam_learning(elevators_domain: Domain) -> SAMLearner:
    return SAMLearner(elevators_domain)


@fixture()
def elevators_sam_learning_soft_policy(elevators_domain: Domain) -> SAMLearner:
    return SAMLearner(elevators_domain, negative_preconditions_policy=NegativePreconditionPolicy.soft)


@fixture()
def elevators_sam_learning_hard_policy(elevators_domain: Domain) -> SAMLearner:
    return SAMLearner(elevators_domain, negative_preconditions_policy=NegativePreconditionPolicy.hard)


@fixture()
def elevators_objects(elevators_observation: Observation) -> Dict[str, PDDLObject]:
    return elevators_observation.grounded_objects


@fixture()
def logistics_domain() -> Domain:
    parser = DomainParser(LOGISTICS_DOMAIN_PATH, partial_parsing=True)
    return parser.parse_domain()


@fixture()
def rovers_domain() -> Domain:
    domain_parser = DomainParser(ROVERS_COMBINED_DOMAIN_PATH, partial_parsing=True)
    return domain_parser.parse_domain()


@fixture()
def woodworking_domain() -> Domain:
    return DomainParser(WOODWORKING_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def woodworking_problem(woodworking_domain: Domain) -> Problem:
    return ProblemParser(problem_path=WOODWORKING_PROBLEM_PATH, domain=woodworking_domain).parse_problem()


@fixture()
def woodworking_observation(woodworking_domain: Domain, woodworking_problem: Problem) -> Observation:
    return TrajectoryParser(woodworking_domain, woodworking_problem).parse_trajectory(WOODWORKING_TRAJECTORY_PATH)


@fixture()
def woodworking_ma_combined_domain() -> Domain:
    return DomainParser(WOODWORKING_COMBINED_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def woodworking_ma_combined_problem(woodworking_ma_combined_domain: Domain) -> Problem:
    return ProblemParser(problem_path=WOODWORKING_COMBINED_PROBLEM_PATH, domain=woodworking_ma_combined_domain).parse_problem()


@fixture()
def multi_agent_observation(woodworking_ma_combined_domain: Domain, woodworking_ma_combined_problem) -> MultiAgentObservation:
    return TrajectoryParser(woodworking_ma_combined_domain, woodworking_ma_combined_problem).parse_trajectory(
        WOODWORKING_COMBINED_TRAJECTORY_PATH, executing_agents=WOODWORKING_AGENT_NAMES
    )


@fixture()
def ma_rovers_domain() -> Domain:
    return DomainParser(ROVERS_COMBINED_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def ma_rovers_problem(ma_rovers_domain: Domain) -> Problem:
    return ProblemParser(problem_path=ROVERS_COMBINED_PROBLEM_PATH, domain=ma_rovers_domain).parse_problem()


@fixture()
def ma_rovers_observation(ma_rovers_domain: Domain, ma_rovers_problem: Problem) -> MultiAgentObservation:
    return TrajectoryParser(ma_rovers_domain, ma_rovers_problem).parse_trajectory(
        ROVERS_COMBINED_TRAJECTORY_PATH, executing_agents=ROVERS_AGENT_NAMES
    )


@fixture()
def spider_domain() -> Domain:
    return DomainParser(SPIDER_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def spider_problem(spider_domain: Domain) -> Problem:
    return ProblemParser(problem_path=SPIDER_PROBLEM_PATH, domain=spider_domain).parse_problem()


@fixture()
def spider_observation(spider_domain: Domain, spider_problem: Problem) -> Observation:
    return TrajectoryParser(spider_domain, spider_problem).parse_trajectory(SPIDER_TRAJECTORY_PATH)


@fixture()
def depot_domain() -> Domain:
    domain_parser = DomainParser(DEPOTS_NUMERIC_DOMAIN_PATH, partial_parsing=True)
    return domain_parser.parse_domain()


@fixture()
def depot_problem(depot_domain: Domain) -> Problem:
    return ProblemParser(problem_path=DEPOTS_NUMERIC_PROBLEM_PATH, domain=depot_domain).parse_problem()


@fixture()
def depot_discrete_domain() -> Domain:
    domain_parser = DomainParser(DEPOTS_DISCRETE_DOMAIN_PATH, partial_parsing=False)
    return domain_parser.parse_domain()


@fixture()
def depot_discrete_problem(depot_discrete_domain: Domain) -> Problem:
    return ProblemParser(problem_path=DEPOTS_DISCRETE_PROBLEM_PATH, domain=depot_discrete_domain).parse_problem()


@fixture()
def depot_observation(depot_domain: Domain, depot_problem: Problem) -> Observation:
    return TrajectoryParser(depot_domain, depot_problem).parse_trajectory(DEPOT_NUMERIC_TRAJECTORY_PATH)


@fixture()
def nurikabe_domain() -> Domain:
    return DomainParser(NURIKABE_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def nurikabe_problem(nurikabe_domain: Domain) -> Problem:
    return ProblemParser(problem_path=NURIKABE_PROBLEM_PATH, domain=nurikabe_domain).parse_problem()


@fixture()
def nurikabe_observation(nurikabe_domain: Domain, nurikabe_problem: Problem) -> Observation:
    return TrajectoryParser(nurikabe_domain, nurikabe_problem).parse_trajectory(NURIKABE_TRAJECTORY_PATH)


@fixture()
def satellite_adl_domain() -> Domain:
    return DomainParser(ADL_SATELLITE_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def satellite_adl_problem(satellite_adl_domain: Domain) -> Problem:
    return ProblemParser(problem_path=ADL_SATELLITE_PROBLEM_PATH, domain=satellite_adl_domain).parse_problem()


@fixture()
def satellite_adl_observation(satellite_adl_domain: Domain, satellite_adl_problem: Problem) -> Observation:
    return TrajectoryParser(satellite_adl_domain, satellite_adl_problem).parse_trajectory(ADL_SATELLITE_TRAJECTORY_PATH)


@fixture()
def minecraft_domain() -> Domain:
    domain_parser = DomainParser(MINECRAFT_DOMAIN_PATH, partial_parsing=True)
    return domain_parser.parse_domain()


@fixture()
def minecraft_full_domain() -> Domain:
    domain_parser = DomainParser(MINECRAFT_DOMAIN_PATH, partial_parsing=False)
    return domain_parser.parse_domain()


@fixture()
def minecraft_problem(minecraft_domain: Domain) -> Problem:
    return ProblemParser(problem_path=MINECRAFT_PROBLEM_PATH, domain=minecraft_domain).parse_problem()


@fixture()
def minecraft_observation(minecraft_domain: Domain, minecraft_problem: Problem) -> Observation:
    return TrajectoryParser(minecraft_domain, minecraft_problem).parse_trajectory(MINECRAFT_TRAJECTORY_PATH)


@fixture()
def satellite_numeric_domain() -> Domain:
    domain_parser = DomainParser(SATELLITE_DOMAIN_PATH, partial_parsing=True)
    return domain_parser.parse_domain()


@fixture()
def satellite_numeric_problem(satellite_numeric_domain: Domain) -> Problem:
    return ProblemParser(problem_path=SATELLITE_PROBLEM_PATH, domain=satellite_numeric_domain).parse_problem()


@fixture()
def satellite_numeric_observation(satellite_numeric_domain: Domain, satellite_numeric_problem: Problem) -> Observation:
    return TrajectoryParser(satellite_numeric_domain, satellite_numeric_problem).parse_trajectory(SATELLITE_NUMERIC_TRAJECTORY_PATH)


@fixture()
def minecraft_large_domain() -> Domain:
    domain_parser = DomainParser(MINECRAFT_LARGE_DOMAIN_PATH, partial_parsing=True)
    return domain_parser.parse_domain()


@fixture()
def minecraft_large_problem(minecraft_large_domain: Domain) -> Problem:
    return ProblemParser(problem_path=MINECRAFT_LARGE_PROBLEM_PATH, domain=minecraft_large_domain).parse_problem()


@fixture()
def minecraft_large_trajectory(minecraft_large_domain: Domain, minecraft_large_problem: Problem) -> Observation:
    return TrajectoryParser(minecraft_large_domain, minecraft_large_problem).parse_trajectory(MINECRAFT_LARGE_TRAJECTORY_PATH)



"""This module contains the LinearRegressionLearner class."""
import logging
from typing import Optional, List, Dict, Tuple, Set

import numpy as np
from pandas import DataFrame
from pddl_plus_parser.models import Precondition, NumericalExpressionTree, PDDLFunction
from sklearn.linear_model import LinearRegression

from sam_learning.core.exceptions import NotSafeActionError
from sam_learning.core.learning_types import EquationSolutionType, ConditionType
from sam_learning.core.numeric_learning.numeric_utils import (
    detect_linear_dependent_features,
    construct_linear_equation_string,
    construct_non_circular_assignment,
    construct_multiplication_strings,
    prettify_coefficients,
    construct_numeric_conditions,
    construct_numeric_effects,
    filter_constant_features,
    get_num_independent_equations,
)

LABEL_COLUMN = "label"
NEXT_STATE_PREFIX = "next_state_"  # prefix for the next state variables.
SUBTRACTION_COLUMN = "subtraction"
POST_NEXT_STATE_PREFIX_INDEX = len(NEXT_STATE_PREFIX)

LEGAL_LEARNING_SCORE = 1.00


class NaiveLinearRegressionLearner:
    def __init__(self, action_name: str, domain_functions: Dict[str, PDDLFunction]):
        self.logger = logging.getLogger(__name__)
        self.action_name = action_name
        self.domain_functions = domain_functions

    @staticmethod
    def _combine_states_data(
        prev_state: Dict[str, List[float]], next_state: Dict[str, List[float]], relevant_fluents: Optional[List[str]] = None
    ) -> DataFrame:
        """Combines the previous and next states data into a single dataframe.

        :param prev_state: the previous state data.
        :param next_state: the next state data.
        :param relevant_fluents: the relevant fluents to use.
        :return: the combined dataframe.
        """
        pre_state_fluents_to_use = (
            [key for key in prev_state.keys() if key in relevant_fluents] if relevant_fluents is not None else list(prev_state.keys())
        )
        next_state_fluents_to_use = (
            [fluent for fluent in next_state.keys() if fluent in relevant_fluents] if relevant_fluents is not None else list(next_state.keys())
        )
        combined_data = {fluent_name: fluent_values for fluent_name, fluent_values in prev_state.items() if fluent_name in pre_state_fluents_to_use}
        combined_data.update(
            {
                f"{NEXT_STATE_PREFIX}{fluent_name}": fluent_values
                for fluent_name, fluent_values in next_state.items()
                if fluent_name in next_state_fluents_to_use
            }
        )
        dataframe = DataFrame(combined_data).fillna(0)
        dataframe.drop_duplicates(inplace=True)
        return dataframe

    def _validate_legal_equations(self, values_df: DataFrame) -> bool:
        """Validates that there are enough independent equations which enable for a single solution for the equation.

        :param values_df: the matrix constructed based on the observations.
        """
        if len(values_df) == 1:
            return False

        num_dimensions = len(values_df.columns.tolist()) + 1  # +1 for the bias.
        num_independent_rows = get_num_independent_equations(values_df)
        if num_independent_rows >= num_dimensions:
            return True

        failure_reason = (
            f"There are too few independent rows of data! " f"cannot create a linear equation from the current data for action - {self.action_name}!"
        )
        self.logger.warning(failure_reason)

        return False

    def _solve_regression_problem(
        self, values_matrix: np.ndarray, function_post_values: np.ndarray, allow_unsafe_learning: bool = True
    ) -> Tuple[List[float], float]:
        """Solves the polynomial equations using a matrix form.

        Note: the equation Ax=b is solved as: x = inverse(A)*b.

        :param values_matrix: the A matrix that contains the previous values of the function variables.
        :param function_post_values: the resulting values after the linear change.
        :param allow_unsafe_learning: whether to allow unsafe learning.
        :return: the vector representing the coefficients for the function variables and the learning score (R^2).
        """
        regressor = LinearRegression()
        regressor.fit(values_matrix, function_post_values)
        learning_score = regressor.score(values_matrix, function_post_values)
        if learning_score < LEGAL_LEARNING_SCORE:
            reason = (
                f"The learned effects are not safe since the R^2 is not high enough. "
                f"Got R^2 of {learning_score} and expected {LEGAL_LEARNING_SCORE}! "
                f"Action {self.action_name} is not safe!"
            )
            self.logger.warning(reason)
            if not allow_unsafe_learning:
                raise NotSafeActionError(self.action_name, reason, EquationSolutionType.no_solution_found)

        coefficients = list(regressor.coef_) + [regressor.intercept_]
        coefficients = prettify_coefficients(coefficients)
        self.logger.debug(f"Learned the coefficients for the numeric equations with r^2 score of {learning_score}")
        return coefficients, learning_score

    def _calculate_scale_factor(self, coefficient_vector: List[float], regression_df: DataFrame, lifted_function: str) -> Optional[float]:
        """

        :param coefficient_vector:
        :param regression_df:
        :param lifted_function:
        :return:
        """
        if (
            lifted_function in regression_df.columns
            and coefficient_vector[list(regression_df.columns).index(lifted_function)] != 0
            and all(
                [
                    coefficient_vector[list(regression_df.columns).index(function)] == 0
                    for function in regression_df.columns
                    if function != lifted_function
                ]
            )
        ):
            self.logger.debug(f"The value of the {lifted_function} is being scaled by a constant factor!")
            return round(coefficient_vector[list(regression_df.columns).index(lifted_function)], 10)

        return None

    def _extract_non_zero_change(self, lifted_function: str, regression_df: DataFrame) -> Tuple[float, float]:
        """

        :param lifted_function:
        :param regression_df:
        :return:
        """
        previous_value = next_value = 0.0
        for i in range(len(regression_df)):
            previous_value = regression_df.iloc[i][lifted_function]
            next_value = regression_df.iloc[i][LABEL_COLUMN]
            if previous_value != next_value:
                break
        return next_value, previous_value

    def _solve_linear_equations(self, lifted_function: str, regression_df: DataFrame, allow_unsafe_learning: bool = False) -> Optional[str]:
        """Computes the change in the function value based on the previous values of the function.

        Note: We assume in this stage that the change results from a polynomial function of the previous values.

        :param lifted_function: the function to compute the change for.
        :param regression_df: the matrix containing the previous values of the function and the resulting values.
        :param allow_unsafe_learning: whether to allow unsafe learning.
        :return: the string representing the polynomial function change in PDDL+ format.
        """
        regression_array = np.array(regression_df.loc[:, regression_df.columns != LABEL_COLUMN])
        function_post_values = np.array(regression_df[LABEL_COLUMN])
        coefficient_vector, learning_score = self._solve_regression_problem(regression_array, function_post_values, allow_unsafe_learning)

        if all([coef == 0 for coef in coefficient_vector]) and len(regression_df[LABEL_COLUMN].unique()) == 1:
            self.logger.debug(
                "The algorithm designated a vector of zeros to the equation "
                "which means that there are no coefficients and the label is also constant value."
            )
            return f"(assign {lifted_function} {regression_df[LABEL_COLUMN].unique()[0]})"

        functions_and_dummy = list(regression_df.columns[:-1]) + ["(dummy)"]
        scale_factor = self._calculate_scale_factor(coefficient_vector, regression_df, lifted_function)
        if scale_factor is not None:
            if scale_factor == 1:
                self.logger.debug("The scale factor is 1, there is no change in the function value.")
                return None

            scale = "scale-up" if scale_factor > 1 else "scale-down"
            self.logger.debug("Currently supporting only scaling of the function by a constant value.")
            return f"({scale} {lifted_function} {scale_factor})"

        if lifted_function in regression_df.columns and coefficient_vector[list(regression_df.columns).index(lifted_function)] != 0:
            self.logger.debug("the function is a part of the equation, cannot use circular format!")
            coefficients_map = {lifted_func: coef for lifted_func, coef in zip(functions_and_dummy, coefficient_vector)}
            next_value, previous_value = self._extract_non_zero_change(lifted_function, regression_df)
            return construct_non_circular_assignment(lifted_function, coefficients_map, previous_value, next_value)

        multiplication_functions = construct_multiplication_strings(coefficient_vector, functions_and_dummy)
        if len(multiplication_functions) == 0:
            self.logger.debug(
                "The algorithm designated a vector of zeros to the equation " "which means that there are not coefficients. Continuing."
            )
            return None

        constructed_right_side = construct_linear_equation_string(multiplication_functions)
        return f"(assign {lifted_function} {constructed_right_side})"

    @staticmethod
    def _extract_const_conditions(precondition_statements: List[List[str]], additional_conditions: List[str] = None) -> List[str]:
        """Extract conditions that are constant for all the samples.

        Note:
            These preconditions that appear in all samples will be considered axioms and will be added the
            general preconditions to create a more compact version of the action's preconditions.

        :param precondition_statements: the precondition statements.
        :return: the extracted constant conditions.
        """
        const_preconditions = set(additional_conditions) if additional_conditions else set()
        for conjunction_statement in precondition_statements:
            for condition in conjunction_statement:
                if all([condition in other_conjunction for other_conjunction in precondition_statements]):
                    const_preconditions.add(condition)
                    continue

        for conjunction_statement in precondition_statements:
            for const_condition in const_preconditions:
                if const_condition in conjunction_statement:
                    conjunction_statement.remove(const_condition)

        return list(const_preconditions)

    def _construct_restrictive_numeric_preconditions(self, combined_data: DataFrame, additional_conditions: List[str] = None) -> Precondition:
        """Constructs the restrictive numeric preconditions for the action.

        Note:
            if there is only one sample we create an AND condition, otherwise we create an OR condition.

        :param combined_data: the combined data of the previous and next states.
        :return: the restrictive numeric preconditions.
        """
        self.logger.debug(f"Constructing the restrictive numeric preconditions for the action {self.action_name}.")
        precondition_statements = []
        for index, row in combined_data.iterrows():
            current_data_conditions = [f"(= {fluent} {row[fluent]})" for fluent in combined_data.columns if not fluent.startswith(NEXT_STATE_PREFIX)]
            precondition_statements.append(current_data_conditions)

        if len(precondition_statements) == 1:
            return construct_numeric_conditions(precondition_statements[0], ConditionType.conjunctive, self.domain_functions)

        const_preconditions = self._extract_const_conditions(precondition_statements, additional_conditions)
        combined_preconditions = None
        if len(const_preconditions) > 0:
            combined_preconditions = construct_numeric_conditions(const_preconditions, ConditionType.conjunctive, self.domain_functions)

        disjunctive_preconditions = Precondition("or")
        for precondition_statement in precondition_statements:
            disjunctive_preconditions.add_condition(
                construct_numeric_conditions(precondition_statement, ConditionType.conjunctive, self.domain_functions)
            )

        if combined_preconditions is not None:
            combined_preconditions.add_condition(disjunctive_preconditions)
            return combined_preconditions

        return disjunctive_preconditions

    def _construct_effect_from_single_sample(
        self, prev_state_data: Dict[str, List[float]], next_state_data: Dict[str, List[float]]
    ) -> Set[NumericalExpressionTree]:
        """constructs the effect from a single sample.

        :return: the constructed numeric effect.
        """
        assignment_statements = []
        for fluent in next_state_data:
            if fluent not in prev_state_data:
                raise ValueError(f"The fluent {fluent} is not in the previous state data.")

            if prev_state_data[fluent][0] != next_state_data[fluent][0]:
                assignment_statements.append(f"(assign {fluent} {next_state_data[fluent][0]})")

        return construct_numeric_effects(assignment_statements, self.domain_functions)

    def _filter_out_constant_effects(self, constant_features: List[str], combined_data_df: DataFrame, filtered_df: DataFrame) -> List[str]:
        """

        :param constant_features:
        :param combined_data_df:
        :param filtered_df:
        :return:
        """
        removed_next_state_variables = []
        for constant_feature in constant_features:
            if combined_data_df[constant_feature].equals(combined_data_df[f"{NEXT_STATE_PREFIX}{constant_feature}"]):
                self.logger.info(
                    f"The value of the {constant_feature} is the same before and "
                    f"after the application of the action. The value is constant so it is not needed!"
                )
                filtered_df.drop(f"{NEXT_STATE_PREFIX}{constant_feature}", axis=1, inplace=True)
                removed_next_state_variables.append(constant_feature)

        return removed_next_state_variables

    def construct_assignment_equations(
        self,
        previous_state_data: Dict[str, List[float]],
        next_state_data: Dict[str, List[float]],
        allow_unsafe_learning: bool = False,
        relevant_fluents: Optional[List[str]] = None,
    ) -> Tuple[Set[NumericalExpressionTree], Optional[Precondition], bool]:
        """Constructs the assignment statements for the action according to the changed value functions.

        :param previous_state_data: the data of the previous state.
        :param next_state_data: the data of the next state.
        :param allow_unsafe_learning: whether to allow unsafe learning.
        :param relevant_fluents: the fluents that are part of the action's preconditions and effects.
        :return: the constructed effects with the possibly additional preconditions.
        """
        if relevant_fluents is not None and len(relevant_fluents) == 0:
            self.logger.debug(f"No numeric effects for the action {self.action_name}.")
            return set(), None, False

        assignment_statements = []
        relevant_next_state_fluents = [fluent_name for fluent_name in next_state_data.keys() if fluent_name in relevant_fluents] if relevant_fluents else list(next_state_data.keys())
        self.logger.info(f"Constructing the fluent assignment equations for action {self.action_name}.")
        combined_data = self._combine_states_data(previous_state_data, next_state_data, relevant_fluents)
        if combined_data.shape[0] == 1:
            raise NotSafeActionError(
                name=self.action_name, reason="Not enough data to learn the numeric effects!", solution_type=EquationSolutionType.not_enough_data
            )

        # The dataset contains more than one observation.
        self.logger.debug("Removing fluents that are constant zero...")
        tagged_next_state_fluents = [f"{NEXT_STATE_PREFIX}{fluent_name}" for fluent_name in relevant_next_state_fluents]
        filtered_df, constant_features_conditions, constant_features = filter_constant_features(
            combined_data, columns_to_ignore=tagged_next_state_fluents
        )
        features_df = filtered_df.copy()[[k for k in previous_state_data.keys() if k in filtered_df.columns]]
        regression_df, linear_dep_conditions, dependent_columns = detect_linear_dependent_features(features_df)
        combined_conditions = linear_dep_conditions + constant_features_conditions
        tested_fluents_names = [fluent_name for fluent_name in relevant_next_state_fluents]
        is_safe_to_learn = self._validate_legal_equations(regression_df)
        for feature_fluent, tagged_fluent in zip(tested_fluents_names, tagged_next_state_fluents):
            regression_df[LABEL_COLUMN] = combined_data[tagged_fluent]
            if combined_data[feature_fluent].equals(combined_data[tagged_fluent]) and is_safe_to_learn:
                self.logger.debug(
                    f"The value of the {feature_fluent} is the same before and "
                    f"after the application of the action. The action does not change the value!"
                )
                continue

            polynomial_equation = self._solve_linear_equations(feature_fluent, regression_df, allow_unsafe_learning=allow_unsafe_learning)

            if polynomial_equation is not None:
                assignment_statements.append(polynomial_equation)

        if not is_safe_to_learn:
            raise NotSafeActionError(
                name=self.action_name, reason="Not enough data to learn the numeric effects!", solution_type=EquationSolutionType.not_enough_data
            )

        self.logger.info(f"Finished constructing the assignment statements for the action {self.action_name}.")
        return (
            construct_numeric_effects(assignment_statements, self.domain_functions),
            construct_numeric_conditions(combined_conditions, ConditionType.conjunctive, self.domain_functions),
            True,
        )



"""Module for exceptions that will be used in the algorithms."""
from sam_learning.core.learning_types import EquationSolutionType


class NotSafeActionError(Exception):
    """An exception type indicating that the learning process is not safe.

    Note:
        The exception contains the reason for the action not being safe to learn. Can be used in the reports.
    """
    action_name: str
    reason: str
    solution_type: EquationSolutionType

    def __init__(self, name: str, reason: str, solution_type: EquationSolutionType):
        self.action_name = name
        self.reason = reason
        self.solution_type = solution_type

    def __str__(self):
        return f"The action {self.action_name} is not safe to use! The reason - {self.reason}"



from .convex_hull_learner import ConvexHullLearner
from .numeric_fluent_learner_algorithm import NumericFluentStateStorage
from .numeric_utils import (
    construct_multiplication_strings,
    prettify_coefficients,
    prettify_floating_point_number,
    construct_linear_equation_string,
    construct_non_circular_assignment,
    divide_span_by_common_denominator,
)



"""Module to create the dataset for the experiments."""
import argparse
import logging
from pathlib import Path
from typing import Literal, Union

from solvers import MetricFFSolver, ENHSPSolver
from trajectory_creators import ExperimentTrajectoriesCreator
from trajectory_creators.random_walk_trajectories_creator import RandomWalkTrajectoriesCreator
from utilities import SolverType

MAX_RUNNING_TIME = 600  # seconds

SOLVER_TYPES = {
    SolverType.metric_ff: MetricFFSolver,
    SolverType.enhsp: ENHSPSolver,
}

TRAJECTORY_CREATOR_TYPES = {
    "regular": ExperimentTrajectoriesCreator,
    "random": RandomWalkTrajectoriesCreator,
}


class DatasetGenerator:
    """Creates the dataset for the experiments.

    Notice: The class assumes that the problems for the experiments are already created (using some sort of generator).
    """

    problems_directory_path: Path
    domain_file_path: Path
    solving_timeout: int
    problems_prefix: str
    solver_type: SolverType
    trajectory_generator: Union[ExperimentTrajectoriesCreator, RandomWalkTrajectoriesCreator]

    def __init__(
        self,
        problems_directory_path: Path,
        domain_file_path: Path,
        solving_timeout: int = MAX_RUNNING_TIME,
        problems_prefix: str = "pfile",
        solver_type: SolverType = SolverType.fast_downward,
        trajectory_creation_method: Literal["regular", "random"] = "regular",
    ):
        """Initializes the dataset generator with the given problems."""
        self.problems_directory_path = problems_directory_path
        self.domain_file_path = domain_file_path
        self.solving_timeout = solving_timeout
        self.problems_prefix = problems_prefix
        self.solver_type = SolverType(solver_type)
        self.trajectory_generator = TRAJECTORY_CREATOR_TYPES[trajectory_creation_method](
            domain_file_name=domain_file_path.name, working_directory_path=problems_directory_path
        )
        self.logger = logging.getLogger("experiments_dataset_generator")

    def generate_dataset(self):
        """Generates the dataset for the experiments."""
        self.logger.info(f"Generating the experimental dataset for the domain {self.domain_file_path.stem}...")
        self.logger.info(f"Using the solver {self.solver_type.name}...")
        solver = SOLVER_TYPES[self.solver_type]()
        solver.execute_solver(
            problems_directory_path=self.problems_directory_path,
            domain_file_path=self.domain_file_path,
            solving_timeout=self.solving_timeout,
            problems_prefix=self.problems_prefix,
        )
        self.logger.info("Finished executing the solver on the problems' dataset.")
        self.logger.info("Starting to generate the trajectories...")
        self.trajectory_generator.create_domain_trajectories()
        self.logger.info("Finished generating the trajectories.")


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Runs the code responsible to create the trajectory dataset for the experiments.")
    parser.add_argument("--problems_directory_path", required=True, help="The path to the directory where the domain and problems are located.")
    parser.add_argument("--domain_file_path", required=True, help="the path to the domain file.")
    parser.add_argument(
        "--solving_timeout", required=False, type=int, help="The timeout to use when running the solver algorithm.", default=60,
    )
    parser.add_argument(
        "--solver_type",
        required=False,
        type=int,
        choices=[2, 3],
        help="The solver that should be used for the sake of validation.\nMetric-FF - 2, ENHSP - 3.",
        default=3,
    )
    parser.add_argument("--problems_prefix", required=False, help="The prefix of the problems' file names", type=str, default="pfile")
    parser.add_argument(
        "--trajectory_creation_method",
        required=False,
        help="The type of trajectories to create for the experiments",
        type=str,
        choices=["regular", "random"],
        default="regular",
    )
    args = parser.parse_args()
    return args


if __name__ == "__main__":
    logging.basicConfig(datefmt="%Y-%m-%d %H:%M:%S", level=logging.INFO)
    args = parse_arguments()
    dataset_generator = DatasetGenerator(
        problems_directory_path=Path(args.problems_directory_path),
        domain_file_path=Path(args.domain_file_path),
        solving_timeout=args.solving_timeout,
        problems_prefix=args.problems_prefix,
        solver_type=args.solver_type,
        trajectory_creation_method=args.trajectory_creation_method,
    )
    dataset_generator.generate_dataset()



import sys

# TODO: replace this temporary patch with update pypi N-SAM package
# Load the actual SAM package (relative import)
from . import utilities as _utilities
from . import sam_learning as _sam_learning

# Inject it into sys.modules under the expected top-level name
sys.modules["utilities"] = _utilities
sys.modules["sam_learning"] = _sam_learning



"""This module contains the ConvexHullLearner class."""
import logging
import os
from pathlib import Path
from typing import Dict, List, Tuple, Optional

import numpy as np
import pandas as pd
from pandas import DataFrame, Series
from pddl_plus_parser.models import Precondition, PDDLFunction
from scipy.spatial import ConvexHull, QhullError

from sam_learning.core.exceptions import NotSafeActionError
from sam_learning.core.learning_types import ConditionType, EquationSolutionType
from sam_learning.core.numeric_learning.numeric_utils import (
    prettify_coefficients,
    construct_numeric_conditions,
    construct_projected_variable_strings,
    extended_gram_schmidt,
    display_convex_hull,
    create_monomials,
    create_polynomial_string,
    divide_span_by_common_denominator,
    remove_complex_linear_dependencies,
    construct_pddl_inequality_scheme,
)

np.set_printoptions(precision=2)


class ConvexHullLearner:
    """Class that learns the convex hull of the preconditions of an action."""

    logger: logging.Logger
    action_name: str
    domain_functions: Dict[str, PDDLFunction]
    convex_hull_error_file_path: Path
    data: DataFrame

    def __init__(
        self, action_name: str, domain_functions: Dict[str, PDDLFunction], polynom_degree: int = 0, epsilon: float = 0.0, qhull_options: str = ""
    ):
        self.logger = logging.getLogger(__name__)
        self.convex_hull_error_file_path = Path(os.environ["CONVEX_HULL_ERROR_PATH"])
        functions = list([function.untyped_representation for function in domain_functions.values()])
        monomials = create_monomials(functions, polynom_degree)
        self.data = DataFrame(columns=[create_polynomial_string(monomial) for monomial in monomials])
        self.action_name = action_name
        self.domain_functions = {function.name: function for function in domain_functions.values()}
        self._epsilon = epsilon
        self._qhull_options = qhull_options

    def add_new_point(self, point: Dict[str, float]) -> None:
        """Adds a new point to the convex hull learner.

        Note:
            This method is supposed to improve the performance of the CH calculations by incrementally adding points.

        :param point: the point to add to the convex hull learner.
        """
        new_sample = DataFrame.from_dict(data={key: [value] for key, value in point.items()}, orient="columns")
        # assuming that if a feature is relevant to the preconditions it should always appear in the dataframe.
        if len(self.data) == 0:
            self.data = new_sample  # This is to avoid observing warnings when adding the first sample.
            self.logger.debug("Added the first sample to the points dataframe.")
            return

        concat_data = pd.concat([self.data, new_sample], ignore_index=True).dropna(axis=1)
        if concat_data.drop_duplicates().shape[0] == self.data.shape[0]:
            self.logger.debug("The new point is already in the storage, not adding it again.")
            return

        self.data = concat_data

    def _epsilon_approximate_hull(self, points: np.ndarray, incremental: bool = False) -> ConvexHull:
        """Approximates the convex hull of the given points with a margin of epsilon and a set of options for the qhull algorithm..

        :param points: The points comprising the convex hull.
        :param incremental: whether to use the incremental version of the qhull algorithm.
        :return: the approximated convex hull.
        """
        self.logger.debug(f"Approximating the convex hull with epsilon {self._epsilon}.")
        # Step 1: Compute the original convex hull
        hull = ConvexHull(points, incremental=incremental)
        if self._epsilon == 0.0 and self._qhull_options == "":
            return hull

        # Step 2: Find the centroid of the original hull (for expanding outward)
        centroid = np.mean(points[hull.vertices], axis=0)

        # Step 3: Calculate new points on expanded facets to include epsilon margin
        expanded_points = []
        for simplex in hull.simplices:
            # Get the points of the simplex (facet)
            facet_points = points[simplex]
            # Calculate the direction vectors from the centroid to each facet point
            directions = facet_points - centroid
            # Normalize directions and expand each point by (1 + epsilon)
            expanded_facet = centroid + (1 + self._epsilon) * directions
            expanded_points.extend(expanded_facet)

        # Combine original points with expanded points
        all_points = np.vstack([points, expanded_points])

        # Step 4: Compute the approximated convex hull on the expanded point set
        approx_hull = ConvexHull(all_points, qhull_options=self._qhull_options, incremental=incremental)
        return approx_hull

    def _execute_convex_hull(self, points: np.ndarray, display_mode: bool = True) -> Tuple[List[List[float]], List[float]]:
        """Runs the convex hull algorithm on the given input points.

        :param points: the points to run the convex hull algorithm on.
        :param display_mode: whether to display the convex hull.
        :return: the coefficients of the planes that represent the convex hull and the border point.
        """
        hull = self._epsilon_approximate_hull(points)
        display_convex_hull(self.action_name, display_mode, hull)
        equations = np.unique(hull.equations, axis=0)

        A = equations[:, : points.shape[1]]
        b = -equations[:, points.shape[1]]
        coefficients = [prettify_coefficients(row) for row in A]
        border_point = prettify_coefficients(b)
        return coefficients, border_point

    def _create_convex_hull_linear_inequalities(
        self, points_df: DataFrame, display_mode: bool = True
    ) -> Tuple[List[List[float]], List[float], List[str], Optional[List[str]]]:
        """Create the convex hull and returns the matrix representing the inequalities.

        :param points_df: the dataframe containing the points that represent the values of the function in the states
            of the observations prior to the action's execution.
        :return: the matrix representing the inequalities of the planes created by the convex hull as well as the
            names of the features that are part of the convex hull.

        Note: the returned values represents the linear inequalities of the convex hull, i.e.,  Ax <= b.
        """
        self.logger.debug(f"Creating convex hull for action {self.action_name}.")
        points = points_df.to_numpy()
        shift_axis = points[0].tolist()  # selected the first vector to be the start of the axis.
        shifted_points = points - shift_axis
        self.logger.debug("Finding the basis vectors for the projected CH using the extended Gram-Schmidt method.")
        projection_basis = extended_gram_schmidt(shifted_points)
        if len(shifted_points) > len(points_df.columns.tolist()) and len(projection_basis) >= len(points_df.columns.tolist()):
            self.logger.debug("The points are spanning the original space and the basis is full rank, " "no need to project the points.")
            coefficients, border_point = self._execute_convex_hull(points, display_mode)
            return coefficients, border_point, points_df.columns.tolist(), []

        projected_points = np.dot(shifted_points, np.array(projection_basis).T)
        if projected_points.shape[1] == 1:
            border_point, coefficients = self._construct_single_dimension_convex_hull(projected_points)

        else:
            coefficients, border_point = self._execute_convex_hull(projected_points, display_mode)

        transformed_vars = construct_projected_variable_strings(points_df.columns.tolist(), shift_axis, projection_basis)

        self.logger.debug("Constructing the conditions to verify that points are in the correct span.")
        diagonal_eye = [list(vector) for vector in np.eye(points.shape[1])]
        orthnormal_span = divide_span_by_common_denominator(extended_gram_schmidt(diagonal_eye, projection_basis))
        transformed_orthonormal_vars = construct_projected_variable_strings(points_df.columns.tolist(), shift_axis, diagonal_eye)
        span_verification_conditions = construct_pddl_inequality_scheme(
            np.array(orthnormal_span), np.zeros(len(orthnormal_span)), transformed_orthonormal_vars, sign_to_use="="
        )
        return coefficients, border_point, transformed_vars, span_verification_conditions

    def _construct_single_dimension_convex_hull(self, projected_points: np.array) -> Tuple[List[float], List[float]]:
        """Construct the convex hull for a single dimension.

        :param projected_points: the projected points on the single dimension.
        :return: the border point and the coefficients of the planes that represent the convex hull.
        """
        self.logger.debug("The convex hull is single dimensional, creating min-max conditions on the new basis.")
        coefficients = [[-1], [1]]
        border_point = prettify_coefficients([-projected_points.min(), projected_points.max()])
        return border_point, coefficients

    def _create_disjunctive_preconditions(self, previous_state_matrix: DataFrame, equality_conditions: List[str] = []) -> Precondition:
        """Create the disjunctive representation of the preconditions.

        :param previous_state_matrix: the matrix containing the previous state values.
        :return: a disjunctive representation for the precondition in case a convex hull cannot be created.
        """
        if previous_state_matrix.shape[0] == 1:
            extended_conditions = [*equality_conditions]
            extended_conditions.extend([f"(= {column} {previous_state_matrix[column].iloc[0]})" for column in previous_state_matrix.columns])
            return construct_numeric_conditions(extended_conditions, condition_type=ConditionType.conjunctive, domain_functions=self.domain_functions)

        disjunctive_precondition = Precondition("or")
        for _, row in previous_state_matrix.iterrows():
            conjunctive_conditions = []
            for column in previous_state_matrix.columns:
                conjunctive_conditions.append(f"(= {column} {row[column]})")

            disjunctive_precondition.add_condition(
                construct_numeric_conditions(
                    conjunctive_conditions, condition_type=ConditionType.conjunctive, domain_functions=self.domain_functions,
                )
            )

        if len(equality_conditions) > 0:
            combined_condition = construct_numeric_conditions(
                equality_conditions, condition_type=ConditionType.conjunctive, domain_functions=self.domain_functions
            )
            combined_condition.add_condition(disjunctive_precondition)
            return combined_condition

        return disjunctive_precondition

    def _construct_single_dimension_inequalities(self, single_column_df: Series) -> Precondition:
        """Construct a single dimension precondition representation.

        :param single_column_df: the fluent only fluent that is relevant to the preconditions' creation.
        :return: the preconditions string and the condition type.
        """
        min_value = single_column_df.min()
        max_value = single_column_df.max()
        relevant_fluent = single_column_df.name
        if min_value == max_value:
            conditions = [f"(= {relevant_fluent} {min_value})"]
        else:
            conditions = [f"(>= {relevant_fluent} {min_value})", f"(<= {relevant_fluent} {max_value})"]

        return construct_numeric_conditions(conditions, condition_type=ConditionType.conjunctive, domain_functions=self.domain_functions)

    def construct_safe_linear_inequalities(self, relevant_fluents: Optional[List[str]] = None) -> Precondition:
        """Constructs the linear inequalities strings that will be used in the learned model later.

        :param relevant_fluents: the relevant fluents that are used to create the convex hull.
        :return: the inequality strings and the type of equations that were constructed (injunctive / disjunctive)
        """
        irrelevant_fluents = (
            [fluent for fluent in self.data.columns.tolist() if fluent not in relevant_fluents] if relevant_fluents is not None else []
        )
        state_data = self.data.drop(columns=irrelevant_fluents)
        if (relevant_fluents is not None and len(relevant_fluents) == 1) or state_data.shape[1] == 1:
            self.logger.debug("Only one dimension is needed in the preconditions!")
            relevant_fluents = state_data.columns.tolist() if relevant_fluents is None else relevant_fluents
            return self._construct_single_dimension_inequalities(state_data.loc[:, relevant_fluents[0]])

        try:
            filtered_dataframe, extra_conditions = remove_complex_linear_dependencies(state_data)
            if filtered_dataframe.shape[0] == 0:
                self.logger.warning("The matrix is empty, no need to create a convex hull.")
                return construct_numeric_conditions(
                    extra_conditions, condition_type=ConditionType.conjunctive, domain_functions=self.domain_functions
                )

            if filtered_dataframe.shape[1] == 1:
                self.logger.debug("After filtering the linear dependencies remained with a single feature!")
                b, A = self._construct_single_dimension_convex_hull(filtered_dataframe.to_numpy())
                inequalities_strs = construct_pddl_inequality_scheme(A, b, filtered_dataframe.columns.tolist())
                return construct_numeric_conditions(
                    [*inequalities_strs, *extra_conditions], condition_type=ConditionType.conjunctive, domain_functions=self.domain_functions
                )

            A, b, column_names, additional_projection_conditions = self._create_convex_hull_linear_inequalities(
                filtered_dataframe, display_mode=False
            )
            inequalities_strs = construct_pddl_inequality_scheme(A, b, column_names)
            if additional_projection_conditions is not None:
                inequalities_strs.extend([*extra_conditions, *additional_projection_conditions])

            return construct_numeric_conditions(inequalities_strs, condition_type=ConditionType.conjunctive, domain_functions=self.domain_functions)

        except (QhullError, ValueError):
            self.logger.warning(
                "Convex hull failed to create a convex hull, using disjunctive preconditions "
                "(probably since the rank of the matrix is 2 and it cannot create a degraded "
                "convex hull)."
            )
            raise NotSafeActionError(
                name=self.action_name, reason="Convex hull failed to create a convex hull", solution_type=EquationSolutionType.convex_hull_not_found
            )



import inspect
from pathlib import Path
import pkgutil
import importlib

registry = {}

package_dir = Path(__file__).resolve().parent
for _, module_name, is_pkg in pkgutil.iter_modules([str(package_dir)]):

    if is_pkg:
        continue
    module = importlib.import_module(f"{__name__}.{module_name}")

    # Expose algorithm classes to allow import such as: from aml_evaluation.algorithms import OffLAM
    if module_name != 'AlgorithmAdapter':
        class_obj = getattr(module, module_name, None)
        assert class_obj is not None, f"{module_name}.{module_name} class is not defined"

        for name, obj in inspect.getmembers(module, inspect.isclass):
            if name.lower() == module_name.lower():
                registry[name] = obj


def get_algorithm(name, **kwargs):
    try:
        return registry[name](**kwargs)
    except KeyError:
        raise ValueError(f"Algorithm '{name}' not found. Available: {list(registry)}")


def print_algorithms() -> None:
    """
    Print available algorithms and their constructor parameters.
    """
    print("Available algorithms:")
    for name, cls in registry.items():
        sig = inspect.signature(cls.__init__)
        params = [
            str(p)
            for pname, p in sig.parameters.items()
            if pname != "self"
            and str(p) != "**kwargs"
        ]
        print(f"  - {name}({', '.join(params)})")



from .online_domain_validator import OnlineLearningDomainValidator
from .safe_domain_validator import DomainValidator
from .validator_script_data import VALIDATOR_DIRECTORY, VALID_PLAN, INAPPLICABLE_PLAN, \
    GOAL_NOT_REACHED, run_validate_script



"""Module containing the datatype of the output domain that the learning algorithms return."""
import os
from collections import defaultdict
from typing import Set, List, Dict

from pddl_plus_parser.models import (
    SignatureType,
    Predicate,
    PDDLType,
    PDDLConstant,
    PDDLFunction,
    Domain,
    ConditionalEffect,
    CompoundPrecondition,
    UniversalEffect,
    NumericalExpressionTree,
)

from utilities.util_types import NUMERIC_PRECISION

DISJUNCTIVE_PRECONDITIONS_REQ = ":disjunctive-preconditions"
NEGATIVE_PRECONDITIONS_REQ = ":negative-preconditions"
EQUALITY_REQ = ":equality"
UNIVERSAL_PRECONDITIONS_REQ = ":universal-preconditions"
CONDITIONAL_EFFECTS_REQ = ":conditional-effects"
ADDED_LEARNING_REQUIREMENTS = [NEGATIVE_PRECONDITIONS_REQ, EQUALITY_REQ]
DEFAULT_DIGITS = int(os.environ.get(NUMERIC_PRECISION, 4))


class LearnerAction:
    """Class representing an action that the learning algorithm outputs."""

    name: str
    signature: SignatureType
    preconditions: CompoundPrecondition
    discrete_effects: Set[Predicate]
    conditional_effects: Set[ConditionalEffect]
    universal_effects: Set[UniversalEffect]
    numeric_effects: Set[NumericalExpressionTree]

    def __init__(self, name: str, signature: SignatureType):
        self.name = name
        self.signature = signature
        self.preconditions = CompoundPrecondition()
        self.discrete_effects = set()
        self.conditional_effects = set()
        self.universal_effects = set()
        self.numeric_effects = set()

    def __str__(self):
        signature_str_items = []
        for parameter_name, parameter_type in self.signature.items():
            signature_str_items.append(f"{parameter_name} - {str(parameter_type)}")

        signature_str = " ".join(signature_str_items)
        return f"({self.name} {signature_str})"

    @property
    def parameter_names(self) -> List[str]:
        return list(self.signature.keys())

    @property
    def preconditions_str_set(self) -> Set[str]:
        """Returns the set of the string representations of the preconditions."""
        precondition_str_set = set()
        for _, precondition in self.preconditions:
            if isinstance(precondition, Predicate):
                precondition_str_set.add(precondition.untyped_representation)

        return precondition_str_set

    def _signature_to_pddl(self) -> str:
        """Converts the action's signature to the PDDL format.

        :return: the PDDL format of the signature.
        """
        signature_str_items = []
        for parameter_name, parameter_type in self.signature.items():
            signature_str_items.append(f"{parameter_name} - {str(parameter_type)}")

        signature_str = " ".join(signature_str_items)
        return f"({signature_str})"

    def _effects_to_pddl(self) -> str:
        """Converts the effects to the needed PDDL format.

        :return: the PDDL format of the effects.
        """
        simple_effects = "\n\t\t".join(sorted([effect.untyped_representation for effect in self.discrete_effects]))

        conditional_effects = "\n\t\t"
        conditional_effects += "\t\t\n".join([str(conditional_effect) for conditional_effect in self.conditional_effects])

        universal_effects = "\n\t\t"
        universal_effects += "\t\t\n".join([str(universal_effect) for universal_effect in self.universal_effects])

        if len(self.numeric_effects) > 0:
            numeric_effects = "\t\t\n".join([effect.to_pddl() for effect in self.numeric_effects])
            return f"(and {simple_effects}\n" f"\t\t{conditional_effects}\n" f"\t\t{universal_effects}\n" f"{numeric_effects})"

        return f"(and {simple_effects} {conditional_effects} {universal_effects})"

    def to_pddl(self, decimal_digits: int = DEFAULT_DIGITS) -> str:
        """Returns the PDDL string representation of the action.

        :param decimal_digits: the number of decimal digits to use to display the preconditions.
        :return: the PDDL string representing the action.
        """
        action_string = (
            f"(:action {self.name}\n"
            f"\t:parameters {self._signature_to_pddl()}\n"
            f"\t:precondition {self.preconditions.print(decimal_digits=decimal_digits)}\n"
            f"\t:effect {self._effects_to_pddl()})"
        )
        formatted_string = "\n".join([line for line in action_string.split("\n") if line.strip()])
        return f"{formatted_string}\n"

    def to_pddl_legacy(self, should_simplify: bool = True, decimal_digits: int = DEFAULT_DIGITS):
        """Returns the PDDL string representation of the action.

        Note:
            This is for legacy support and should be removed in the future.

        :param should_simplify: whether to simplify the preconditions or not.
        :param decimal_digits: the number of decimal digits to use to display the preconditions.
        :return: the PDDL string representing the action.
        """
        action_string = (
            f"(:action {self.name}\n"
            f"\t:parameters {self._signature_to_pddl()}\n"
            f"\t:precondition {self.preconditions.print(should_simplify=should_simplify, decimal_digits=decimal_digits)}\n"
            f"\t:effect {self._effects_to_pddl()})"
        )
        formatted_string = "\n".join([line for line in action_string.split("\n") if line.strip()])
        return f"{formatted_string}\n"


class LearnerDomain:
    """Class representing the domain that is to be learned by the action model learning algorithm."""

    name: str
    requirements: List[str]
    types: Dict[str, PDDLType]
    constants: Dict[str, PDDLConstant]
    predicates: Dict[str, Predicate]
    functions: Dict[str, PDDLFunction]
    actions: Dict[str, LearnerAction]

    # processes: Dict[str, Action] - TBD
    # events: Dict[str, Action] - TBD

    def __init__(self, domain: Domain):
        self.name = domain.name
        self.requirements = domain.requirements
        self.types = domain.types
        self.constants = domain.constants
        self.predicates = domain.predicates
        self.functions = domain.functions
        self.actions = {}
        for action_name, action_object in domain.actions.items():
            self.actions[action_name] = LearnerAction(name=action_name, signature=action_object.signature)

    def __str__(self):
        return "< Domain definition: %s\n Requirements: %s\n Predicates: %s\n Functions: %s\n Actions: %s\n " "Constants: %s >" % (
            self.name,
            [req for req in self.requirements],
            [str(p) for p in self.predicates.values()],
            [str(f) for f in self.functions.values()],
            [str(a) for a in self.actions.values()],
            [str(c) for c in self.constants],
        )

    def _complete_missing_requirements(self) -> None:
        """Completes the requirements of the domain from the needed requirements of the learning algorithm."""
        for requirement in ADDED_LEARNING_REQUIREMENTS:
            if requirement not in self.requirements:
                self.requirements.append(requirement)

    def _types_to_pddl(self) -> str:
        """Converts the types to a PDDL string.

        :return: the PDDL string representing the types.
        """
        parent_child_map = defaultdict(list)
        if set(self.types.keys()) == {"object"}:
            return "object"

        for type_name, type_obj in self.types.items():
            if type_name == "object":
                continue

            parent_child_map[type_obj.parent.name].append(type_name)

        types_strs = []
        for parent_type, children_types in parent_child_map.items():
            types_strs.append(f"\t{' '.join(children_types)} - {parent_type}")

        return "\n".join(types_strs)

    def _constants_to_pddl(self) -> str:
        """Converts the constants to a PDDL string.

        :return: the PDDL string representing the constants.
        """
        same_type_constant = defaultdict(list)
        for const_name, constant in self.constants.items():
            if const_name == "object":
                continue

            same_type_constant[constant.type.name].append(const_name)

        types_strs = []
        for constant_type_name, constant_objects in same_type_constant.items():
            types_strs.append(f"\t{' '.join(constant_objects)} - {constant_type_name}")

        return "\n".join(types_strs)

    def _functions_to_pddl(self) -> str:
        """Converts the functions to PDDL format.

        :return: the PDDL format of the functions.
        """
        return "\n\t".join([str(f) for f in self.functions.values()])

    def to_pddl(self, should_simplify: bool = True, decimal_digits: int = DEFAULT_DIGITS) -> str:
        """Converts the domain into a PDDL string format.

        :return: the PDDL string representing the domain.
        """
        # TODO: remove the support for legacy PDDL format in the future.
        self._complete_missing_requirements()
        predicates = "\n\t".join([str(p) for p in self.predicates.values()])
        predicates_str = f"(:predicates {predicates}\n)\n\n" if len(self.predicates) > 0 else ""
        types_str = f"(:types {self._types_to_pddl()}\n)\n\n" if len(self.types) > 0 else ""
        actions = "\n".join(action.to_pddl_legacy(should_simplify=should_simplify, decimal_digits=decimal_digits) for action in self.actions.values())
        constants = f"(:constants {self._constants_to_pddl()}\n)\n\n" if len(self.constants) > 0 else ""
        functions = f"(:functions {self._functions_to_pddl()}\n)\n\n" if len(self.functions) > 0 else ""
        return (
            f"(define (domain {self.name})\n"
            f"(:requirements {' '.join(self.requirements)})\n"
            f"{types_str}"
            f"{constants}"
            f"{predicates_str}"
            f"{functions}"
            f"{actions}\n)"
        )



import argparse
import logging
from pathlib import Path
from typing import List

from pddl_plus_parser.lisp_parsers import DomainParser

from experiments.concurrent_execution.distributed_results_collector import DistributedResultsCollector
from experiments.plotting.plot_nsam_results import plot_results
from experiments.plotting.plot_nsam_solo_results import plot_solo_results


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Creates the folds directories for the slurm tasks to work on.")
    parser.add_argument("--working_directory_path", required=True, help="The path to the directory where the domain is")
    parser.add_argument("--domain_file_name", required=True, help="the domain file name including the extension")
    parser.add_argument("--learning_algorithms", required=True, help="the list of algorithms that will run in parallel")
    parser.add_argument("--num_folds", required=True, help="the number of folds to that were created", type=int, default=5)
    parser.add_argument("--internal_iterations", required=True, help="The internal iterations that the algorithm will run in parallel.")
    args = parser.parse_args()
    return args


class NumericResultsCollector(DistributedResultsCollector):
    logger: logging.Logger

    def __init__(
        self,
        working_directory_path: Path,
        domain_file_name: str,
        learning_algorithms: List[str] = None,
        num_folds: int = 5,
        iterations: List[int] = None,
    ):
        super().__init__(working_directory_path, domain_file_name, learning_algorithms, num_folds, iterations)

    def collect_numeric_statistics(self) -> None:
        """Collects the statistics from the results directory."""
        self._collect_solving_statistics(collecting_triplets=False)
        results_directory = self.working_directory_path / "results_directory"
        plot_results(results_directory, file_template="*solving_combined_statistics.csv")
        plot_solo_results(results_directory, file_template="*solving_combined_statistics.csv")
        domain = DomainParser(self.working_directory_path / self.domain_file_name).parse_domain()
        self._collect_syntactic_performance_statistics(domain)
        self._collect_numeric_semantic_performance_statistics(collecting_triplets=False)

    def collect_numeric_statistics_with_triplets_statistics(self) -> None:
        self._collect_solving_statistics(collecting_triplets=True)
        results_directory = self.working_directory_path / "results_directory"
        plot_results(results_directory, file_template="*solving_combined_statistics_with_triplets.csv", using_triplets=True)
        plot_solo_results(results_directory, file_template="*solving_combined_statistics_with_triplets.csv", using_triplets=True)
        domain = DomainParser(self.working_directory_path / self.domain_file_name).parse_domain()
        self._collect_syntactic_performance_statistics(domain)
        self._collect_numeric_semantic_performance_statistics(collecting_triplets=True)


if __name__ == "__main__":
    logging.basicConfig(format="%(asctime)s %(name)s %(levelname)-8s %(message)s", datefmt="%Y-%m-%d %H:%M:%S", level=logging.INFO)
    args = parse_arguments()
    experiment_learning_algorithms = args.learning_algorithms.split(",")
    internal_iterations = [int(val) for val in args.internal_iterations.split(",")]
    NumericResultsCollector(
        working_directory_path=Path(args.working_directory_path),
        domain_file_name=args.domain_file_name,
        learning_algorithms=experiment_learning_algorithms,
        num_folds=args.num_folds,
        iterations=internal_iterations,
    ).collect_numeric_statistics()



"""The Safe Action Model Learning algorithm module."""

import logging
import time
from collections import defaultdict
from itertools import combinations
from typing import List, Tuple, Dict, Set

from pddl_plus_parser.models import (
    Observation,
    Predicate,
    ActionCall,
    State,
    Domain,
    ObservedComponent,
    PDDLObject,
    Action,
)

from sam_learning.core import (
    PredicatesMatcher,
    extract_discrete_effects,
    LearnerDomain,
    contains_duplicates,
    VocabularyCreator,
    EnvironmentSnapshot,
)
from utilities import NegativePreconditionPolicy


class SAMLearner:
    """Class that represents the safe action model learner algorithm.

    Notice: This class does not support domains with constants or with the same object mapped to multiple parameters.
    """

    logger: logging.Logger
    partial_domain: Domain
    matcher: PredicatesMatcher
    observed_actions: List[str]
    safe_actions: List[str]
    triplet_snapshot: EnvironmentSnapshot
    current_trajectory_objects: Dict[str, PDDLObject]
    learning_start_time: float
    learning_end_time: float
    vocabulary_creator: VocabularyCreator
    cannot_be_effect: Dict[str, Set[Predicate]]
    negative_preconditions_policy: NegativePreconditionPolicy

    def __init__(
        self,
        partial_domain: Domain,
        negative_preconditions_policy: NegativePreconditionPolicy = NegativePreconditionPolicy.no_remove,
        should_enforce_injective_binding: bool = False,
    ):

        self.logger = logging.getLogger(__name__)
        self.partial_domain = partial_domain
        self.matcher = PredicatesMatcher(partial_domain)
        self.vocabulary_creator = VocabularyCreator()
        self.triplet_snapshot = EnvironmentSnapshot(partial_domain=partial_domain)
        self.observed_actions = []
        self.safe_actions = []
        self.current_trajectory_objects = {}
        self.learning_start_time = 0
        self.learning_end_time = 0
        self.cannot_be_effect = {action: set() for action in self.partial_domain.actions}
        self._action_signatures = {action_name: action.signature for action_name, action in partial_domain.actions.items()}
        self.negative_preconditions_policy = negative_preconditions_policy
        self.should_enforce_injective_binding = should_enforce_injective_binding

    def _handle_unobserved_actions(self):
        """Removes the actions that were not observed from the partial domain."""
        self.logger.debug("Removing unobserved actions from the partial domain")
        unobserved_actions = [action for action in self.partial_domain.actions if action not in self.observed_actions]
        for action in unobserved_actions:
            preconditions_superset = self.vocabulary_creator.create_lifted_vocabulary(
                self.partial_domain, self.partial_domain.actions[action].signature
            )
            for precondition in preconditions_superset:
                self.partial_domain.actions[action].preconditions.add_condition(precondition)

    def _handle_consts_in_effects(self, grounded_action: ActionCall) -> None:
        """The function filters out literals that cannot be effects because they are not observed in the next state.

        :param grounded_action: the grounded action that was executed according to the trajectory.
        """
        action_predicate_vocabulary = self.vocabulary_creator.create_lifted_vocabulary(
            self.partial_domain, self.partial_domain.actions[grounded_action.name].signature
        )
        lifted_next_state_predicates = self.matcher.get_possible_literal_matches(
            grounded_action, list(self.triplet_snapshot.next_state_predicates)
        )
        lifted_next_state_predicates_str = {predicate.untyped_representation for predicate in lifted_next_state_predicates}
        for predicate in action_predicate_vocabulary:
            if predicate.untyped_representation not in lifted_next_state_predicates_str:
                self.cannot_be_effect[grounded_action.name].add(predicate)

        for predicate in self.cannot_be_effect[grounded_action.name]:
            self.partial_domain.actions[grounded_action.name].discrete_effects.discard(predicate)

    def _handle_action_effects(self, grounded_action: ActionCall) -> Tuple[List[Predicate], List[Predicate]]:
        """Finds the effects generated from the previous and the next state on this current step.

        :param grounded_action: the grounded action that was executed according to the trajectory.
        :return: the effect containing the add and del list of predicates.
        """
        self.logger.debug(f"Starting to learn the effects of {str(grounded_action)}.")
        grounded_add_effects, grounded_del_effects = extract_discrete_effects(
            self.triplet_snapshot.previous_state_predicates, self.triplet_snapshot.next_state_predicates
        )
        self.logger.debug("Updating the negative state predicates based on the action's execution.")
        lifted_add_effects = self.matcher.get_possible_literal_matches(grounded_action, list(grounded_add_effects))
        lifted_delete_effects = self.matcher.get_possible_literal_matches(grounded_action, list(grounded_del_effects))
        return lifted_add_effects, lifted_delete_effects

    def _update_action_preconditions(self, grounded_action: ActionCall) -> None:
        """Updates the preconditions of an action after it was observed at least once.

        :param grounded_action: the grounded action that is being executed in the trajectory component.
        """
        current_action = self.partial_domain.actions[grounded_action.name]
        previous_state_predicates = set(
            self.matcher.get_possible_literal_matches(grounded_action, list(self.triplet_snapshot.previous_state_predicates))
        )

        conditions_to_remove = []
        for current_precondition in current_action.preconditions.root.operands:
            # assuming that the predicates in the preconditions are NOT nested.
            if isinstance(current_precondition, Predicate) and current_precondition not in previous_state_predicates:
                conditions_to_remove.append(current_precondition)

        for condition in conditions_to_remove:
            current_action.preconditions.remove_condition(condition)

    def _add_new_action_preconditions(self, grounded_action: ActionCall) -> None:
        """General method to add new action's discrete preconditions.

        :param grounded_action: the action that is currently being executed.
        """
        self.logger.debug(f"Adding the preconditions of {grounded_action.name} to the action model.")
        current_action = self.partial_domain.actions[grounded_action.name]
        previous_state_predicates = set(
            self.matcher.get_possible_literal_matches(grounded_action, list(self.triplet_snapshot.previous_state_predicates))
        )

        for predicate in previous_state_predicates:
            current_action.preconditions.add_condition(predicate)

    def _construct_learning_report(self) -> Dict[str, str]:
        """Constructs the learning report of the learned actions.

        :return: the learning report.
        """
        observed_unsafe_actions = set(self.observed_actions).difference(self.safe_actions)
        unobserved_actions = set(self.partial_domain.actions.keys()).difference(self.observed_actions)

        learning_report = {action_name: "OK" for action_name in self.safe_actions}
        learning_report.update(
            {action_name: "NOT SAFE" for action_name in self.partial_domain.actions if action_name in observed_unsafe_actions}
        )
        learning_report.update({action_name: "UNOBSERVED" for action_name in unobserved_actions})
        learning_report["learning_time"] = str(self.learning_end_time - self.learning_start_time)
        return learning_report

    def add_new_action(self, grounded_action: ActionCall, previous_state: State, next_state: State) -> None:
        """Create a new action in the domain.

        :param grounded_action: the grounded action that was executed according to the trajectory.
        :param previous_state: the state that the action was executed on.
        :param next_state: the state that was created after executing the action on the previous
        """

        self.logger.info(f"Adding the action {str(grounded_action)} to the domain.")
        # adding the preconditions each predicate is grounded in this stage.
        observed_action = self.partial_domain.actions[grounded_action.name]
        self._add_new_action_preconditions(grounded_action)
        lifted_add_effects, lifted_delete_effects = self._handle_action_effects(grounded_action)

        observed_action.discrete_effects.update(set(lifted_add_effects).union(lifted_delete_effects))
        self.observed_actions.append(observed_action.name)
        self.logger.debug(f"Finished adding the action {grounded_action.name}.")

    def update_action(self, grounded_action: ActionCall, previous_state: State, next_state: State) -> None:
        """Create a new action in the domain.

        :param grounded_action: the grounded action that was executed according to the trajectory.
        :param previous_state: the state that the action was executed on.
        :param next_state: the state that was created after executing the action on the previous
            state.
        """
        action_name = grounded_action.name
        observed_action = self.partial_domain.actions[action_name]
        self._update_action_preconditions(grounded_action)
        lifted_add_effects, lifted_delete_effects = self._handle_action_effects(grounded_action)

        observed_action.discrete_effects.update(set(lifted_add_effects).union(lifted_delete_effects))
        if len(self.partial_domain.constants) > 0:
            self._handle_consts_in_effects(grounded_action)

        self.logger.debug(f"Done updating the action - {grounded_action.name}")

    def _verify_parameter_duplication(self, grounded_action: ActionCall) -> bool:
        """Verifies if the action was called with duplicated objects in a trajectory component.

        :param grounded_action: the grounded action observed in the trajectory triplet.
        :return: whether the action contains duplicated parameters.
        """
        has_duplicates = contains_duplicates(grounded_action.parameters)
        if has_duplicates:
            action = self.partial_domain.actions[grounded_action.name]
            grounded_signature_map = defaultdict(list)
            for grounded_param, lifted_param in zip(grounded_action.parameters, action.parameter_names):
                grounded_signature_map[grounded_param].append(lifted_param)

            for lifted_duplicates_list in grounded_signature_map.values():
                for obj1, obj2 in combinations(lifted_duplicates_list, 2):
                    action.preconditions.root.inequality_preconditions.discard((obj1, obj2))

        return has_duplicates

    def _complete_possibly_missing_actions(self):
        """Completes the actions that were not observed in the trajectory.

        Note:
            This allows SAM to be used multiple times on different trajectories.
            In case some actions were deleted in previous iterations they will be added back.
        """
        for action_name, action_signature in self._action_signatures.items():
            if action_name not in self.partial_domain.actions:
                action = Action()
                action.name = action_name
                action.signature = action_signature
                self.partial_domain.actions[action_name] = action

    def handle_single_trajectory_component(self, component: ObservedComponent) -> None:
        """Handles a single trajectory component as a part of the learning process.

        :param component: the trajectory component that is being handled at the moment.
        """
        previous_state = component.previous_state
        grounded_action = component.grounded_action_call
        next_state = component.next_state

        if self._verify_parameter_duplication(grounded_action):
            self.logger.warning(
                f"{str(grounded_action)} contains duplicated parameters! Not supported in SAM." f"aborting learning from component"
            )
            return

        self.triplet_snapshot.create_triplet_snapshot(
            previous_state=previous_state,
            next_state=next_state,
            current_action=grounded_action,
            observation_objects=self.current_trajectory_objects,
        )
        if grounded_action.name not in self.observed_actions:
            self.add_new_action(grounded_action, previous_state, next_state)

        else:
            self.update_action(grounded_action, previous_state, next_state)

    def deduce_initial_inequality_preconditions(self) -> None:
        """Tries to deduce which objects in the actions' signature cannot be equal."""
        if not self.should_enforce_injective_binding:
            return

        self.logger.debug("Starting to deduce inequality preconditions")
        for action_name, action_data in self.partial_domain.actions.items():
            for lifted_param1, lifted_param2 in combinations(action_data.parameter_names, 2):
                if action_data.signature[lifted_param1] == action_data.signature[lifted_param2]:
                    action_data.preconditions.root.inequality_preconditions.add((lifted_param1, lifted_param2))

    def remove_negative_preconditions(self):
        """Removes all negative preconditions"""
        for action in self.partial_domain.actions.values():
            new_preconditions = set()

            for precondition in action.preconditions.root.operands:
                if isinstance(precondition, Predicate) and not precondition.is_positive:
                    action_add_effects = [effect.untyped_representation for effect in action.discrete_effects if effect.is_positive]
                    copy_precondition_positive = precondition.copy(is_negated=True)
                    if (
                        not self.negative_preconditions_policy == NegativePreconditionPolicy.soft
                    ) or copy_precondition_positive.untyped_representation in action_add_effects:
                        continue

                new_preconditions.add(precondition)

            action.preconditions.root.operands = new_preconditions

    def handle_negative_preconditions_policy(self):
        if not self.negative_preconditions_policy == NegativePreconditionPolicy.no_remove:
            self.remove_negative_preconditions()

    def construct_safe_actions(self) -> None:
        """Constructs the single-agent actions that are safe to execute."""
        pass

    def start_measure_learning_time(self) -> None:
        """Starts measuring the learning time."""
        self.logger.info("Starting to measure the time it takes to learn the action model!")
        self.learning_start_time = time.time()

    def end_measure_learning_time(self) -> None:
        """Ends measuring the learning time."""
        self.learning_end_time = time.time()
        self.logger.info(f"Finished learning the action model in " f"{self.learning_end_time - self.learning_start_time} seconds.")

    def learn_action_model(self, observations: List[Observation]) -> Tuple[LearnerDomain, Dict[str, str]]:
        """Learn the SAFE action model from the input trajectories.

        :param observations: the list of trajectories that are used to learn the safe action model.
        :return: a domain containing the actions that were learned.
        """
        self.logger.info("Starting to learn the action model!")
        self.start_measure_learning_time()
        self.deduce_initial_inequality_preconditions()
        self._complete_possibly_missing_actions()
        for observation in observations:
            self.current_trajectory_objects = observation.grounded_objects
            for component in observation.components:
                if not component.is_successful:
                    self.logger.warning("Skipping the transition because it was not successful.")
                    continue

                self.handle_single_trajectory_component(component)

        self.construct_safe_actions()
        self._handle_unobserved_actions()
        self.handle_negative_preconditions_policy()
        self.end_measure_learning_time()
        learning_report = self._construct_learning_report()
        return self.partial_domain, learning_report



"""Utilities for the performance calculation process."""

from typing import Dict

from pddl_plus_parser.models import Domain, Operator, ActionCall, State, PDDLObject


def _ground_executed_action(action_call: ActionCall, learned_domain: Domain, problem_objects: Dict[str, PDDLObject]) -> Operator:
    """Ground the tested action based on the trajectory data.

    :param action_call: the grounded action call in the observation component.
    :param learned_domain: the domain that was learned using the action model learning algorithm.
    :param problem_objects: the objects that were used in the problem definition.
    :return: the grounded operator.
    """
    grounded_operator = Operator(
        action=learned_domain.actions[action_call.name],
        domain=learned_domain,
        grounded_action_call=action_call.parameters,
        problem_objects=problem_objects,
    )
    grounded_operator.ground()
    return grounded_operator


def _calculate_single_action_applicability_rate(
    action_call: ActionCall,
    learned_domain: Domain,
    complete_domain: Domain,
    num_false_negatives: Dict[str, int],
    num_false_positives: Dict[str, int],
    num_true_positives: Dict[str, int],
    observed_state: State,
    problem_objects: Dict[str, PDDLObject],
):
    """Calculates the applicability rate for a single action.

    :param action_call: the grounded action that is currently being tested.
    :param learned_domain: the domain that was learned using the action model learning algorithm.
    :param complete_domain: the complete domain that was used to generate the trajectory data.
    :param num_false_negatives: the number of false negatives for each action.
    :param num_false_positives: the number of false positives for each action.
    :param num_true_positives: the number of true positives for each action.
    :param observed_state: the state that was observed in the trajectory data.
    :param problem_objects: the objects that were used in the problem definition.
    """
    learned_grounded_action = _ground_executed_action(action_call, learned_domain, problem_objects)
    model_grounded_action = _ground_executed_action(action_call, complete_domain, problem_objects)
    is_applicable_in_test = learned_grounded_action.is_applicable(observed_state)
    is_applicable_in_model = model_grounded_action.is_applicable(observed_state)
    num_true_positives[action_call.name] += int(is_applicable_in_test == is_applicable_in_model and is_applicable_in_test)
    num_false_positives[action_call.name] += int(is_applicable_in_test and not is_applicable_in_model)
    num_false_negatives[action_call.name] += int(not is_applicable_in_test and is_applicable_in_model)



import argparse
import logging
import shutil
from pathlib import Path
from typing import List

from trajectory_creators.random_walk_trajectories_creator import RandomWalkTrajectoriesCreator
from utilities import DistributedKFoldSplit

DEFAULT_EXPERIMENT_SIZE = 100


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Creates the folds directories for the slurm tasks to work on.")
    parser.add_argument("--working_directory_path", required=True, help="The path to the directory where the domain is")
    parser.add_argument("--domain_file_name", required=True, help="the domain file name including the extension")
    parser.add_argument(
        "--problem_prefix", required=True, help="the prefix of the name of the problems to be able to generate trajectories from them."
    )
    parser.add_argument("--learning_algorithms", required=True, help="the list of algorithms that will run in parallel")
    parser.add_argument("--internal_iterations", required=True, help="The internal iterations that the algorithm will run in parallel.")
    parser.add_argument(
        "--experiment_size", required=False, help="The max items to use in each fold of the experiment.", default=DEFAULT_EXPERIMENT_SIZE, type=int
    )
    parser.add_argument(
        "--no_random_trajectories",
        action="store_false",
        dest="should_create_random_trajectories",
        help="If specified, no random trajectories would be created.",
    )
    parsed_args = parser.parse_args()
    return parsed_args


class FoldsCreator:
    k_fold: DistributedKFoldSplit
    logger: logging.Logger

    def __init__(
        self, working_directory_path: Path, domain_file_name: str, learning_algorithms: List[int] = None, internal_iterations: List[int] = None
    ):
        self.k_fold = DistributedKFoldSplit(
            working_directory_path=working_directory_path,
            domain_file_name=domain_file_name,
            n_split=5,
            learning_algorithms=learning_algorithms,
            internal_iterations=internal_iterations,
        )
        self.domain_file_name = domain_file_name
        self.working_directory_path = working_directory_path
        self.logger = logging.getLogger("cluster-folder-creator")

    def create_folds_from_cross_validation(self, experiment_size: int = DEFAULT_EXPERIMENT_SIZE) -> None:
        """Runs that cross validation process on the domain's working directory and validates the results."""
        (self.working_directory_path / "results_directory").mkdir(exist_ok=True)
        self.logger.info("Removing the old folds directories if exist.")
        self.k_fold.remove_created_directories()
        self.logger.info("Done removing the old folds directories!")
        self.logger.info("Creating a logs directory.")
        (self.working_directory_path / "logs").mkdir(exist_ok=True)
        self.logger.info("Creating the folds directories.")
        self.k_fold.create_k_fold(max_items=experiment_size)
        self.logger.info("Done creating the folds directories!")

    def create_random_performance_evaluation_trajectories(
        self, problem_prefix: str, learning_algorithm: int, should_include_random_trajectories: bool = True
    ) -> None:
        """Creates the random performance evaluation trajectories.

        :param problem_prefix: the prefix of the name of the problems to be able to generate trajectories from them.
        :param learning_algorithm: the learning algorithm that will be used to create the trajectories (index).
        :param should_include_random_trajectories: whether to include the random trajectories or not.
        """
        self.logger.info("Creating the random performance evaluation trajectories.")
        performance_evaluation_dir = self.working_directory_path / "performance_evaluation_trajectories"
        performance_evaluation_dir.mkdir(exist_ok=True)

        for fold in range(self.k_fold.n_split):
            output_directory_path = performance_evaluation_dir / f"fold_{fold}"
            output_directory_path.mkdir(exist_ok=True)
            # this is the directory containing the test samples that will be used to create the random trajectories
            test_set_directory = self.k_fold.test_set_dir_path / f"fold_{fold}_{learning_algorithm}_1"
            shutil.copy(self.working_directory_path / self.domain_file_name, test_set_directory)
            if should_include_random_trajectories:
                random_trajectories_creator = RandomWalkTrajectoriesCreator(
                    domain_file_name=self.domain_file_name, working_directory_path=test_set_directory
                )
                random_trajectories_creator.create_domain_trajectories(problems_prefix=problem_prefix, output_directory_path=output_directory_path)

            # copy the test set problems, solution and trajectories to use in case none of the random trajectories are valid
            for test_set_problem in test_set_directory.glob(f"{problem_prefix}*.pddl"):
                shutil.copy(test_set_problem, output_directory_path)
                shutil.copy(self.working_directory_path / f"{test_set_problem.stem}.solution", output_directory_path)
                shutil.copy(self.working_directory_path / f"{test_set_problem.stem}.trajectory", output_directory_path)

            (test_set_directory / self.domain_file_name).unlink()

        self.logger.info("Done creating the random performance evaluation trajectories!")


if __name__ == "__main__":
    logging.basicConfig(format="%(asctime)s %(name)s %(levelname)-8s %(message)s", datefmt="%Y-%m-%d %H:%M:%S", level=logging.INFO)
    args = parse_arguments()
    experiment_learning_algorithms = args.learning_algorithms.split(",")
    internal_iterations = [int(val) for val in args.internal_iterations.split(",")]
    if len(internal_iterations) > 0:
        print(f"Internal iterations: {internal_iterations}")

    folds_creator = FoldsCreator(
        working_directory_path=Path(args.working_directory_path),
        domain_file_name=args.domain_file_name,
        learning_algorithms=experiment_learning_algorithms,
        internal_iterations=internal_iterations,
    )
    folds_creator.create_folds_from_cross_validation(experiment_size=args.experiment_size)
    folds_creator.create_random_performance_evaluation_trajectories(
        problem_prefix=args.problem_prefix,
        learning_algorithm=int(experiment_learning_algorithms[0]),
        should_include_random_trajectories=args.should_create_random_trajectories,
    )



"""Extension to SAM Learning that can learn numeric state variables."""

from typing import List, Dict, Optional

from pddl_plus_parser.models import ActionCall, State, Domain

from sam_learning.core import NumericFunctionMatcher
from sam_learning.core.numeric_learning import NumericFluentStateStorage
from sam_learning.learners.numeric_sam import NumericSAMLearner
from utilities import NegativePreconditionPolicy


class IncrementalNumericSAMLearner(NumericSAMLearner):
    """The Extension of SAM that is able to learn numeric state variables."""

    storage: Dict[str, NumericFluentStateStorage]
    function_matcher: NumericFunctionMatcher
    preconditions_fluent_map: Dict[str, List[str]]

    def __init__(
        self,
        partial_domain: Domain,
        polynomial_degree: int = 0,
        allow_unsafe: bool = False,
        negative_preconditions_policy: NegativePreconditionPolicy = NegativePreconditionPolicy.soft,
        epsilon: float = 0.0,
        qhull_options: str = "",
        relevant_fluents: Optional[Dict[str, List[str]]] = None,
    ):
        super().__init__(
            partial_domain,
            negative_preconditions_policy=negative_preconditions_policy,
            polynomial_degree=polynomial_degree,
            relevant_fluents=relevant_fluents,
            epsilon=epsilon,
            qhull_options=qhull_options,
            allow_unsafe=allow_unsafe,
        )

    def add_new_action(self, grounded_action: ActionCall, previous_state: State, next_state: State) -> None:
        """Adds a new action to the learned domain.

        :param grounded_action: the grounded action that was executed according to the observation.
        :param previous_state: the state that the action was executed on.
        :param next_state: the state that was created after executing the action on the previous
        """
        super().add_new_action(grounded_action, previous_state, next_state, incremental=True)



import pathlib
import signal
import time
from datetime import datetime

from experiments.cluster_scripts.common import (
    submit_job,
    progress_bar,
    sigint_handler,
    get_configurations,
    get_environment_variables,
    validate_job_running,
    create_all_experiments_folders,
    EXPERIMENTS_CONFIG_STR,
    submit_job_and_validate_execution,
    create_execution_arguments,
)

signal.signal(signal.SIGINT, sigint_handler)
learning_algorithms_map = {3: "nsam", 15: "naive_nsam", 5: "plan_miner"}


def execute_statistics_collection_job(code_directory, configuration, environment_variables, experiment, job_ids, internal_iterations):
    print(f"Creating the job that will collect the statistics from all the domain's experiments.")
    filtered_sids = [sid for sid in job_ids if validate_job_running(sid) is not None]
    statistics_collection_job = submit_job(
        conda_env="online_nsam",
        mem="4G",
        python_file=f"{code_directory}/numeric_distributed_results_collector.py",
        dependency=f"afterok:{':'.join([str(e) for e in filtered_sids])}",
        jobname=f"collect_statistics_{experiment['domain_file_name']}",
        suppress_output=False,
        arguments=[
            f"--working_directory_path {experiment['working_directory_path']}",
            f"--domain_file_name {experiment['domain_file_name']}",
            f"--learning_algorithms {','.join([str(e) for e in experiment['compared_versions']])}",
            f"--num_folds {configuration['num_folds']}",
            f"--internal_iterations {','.join([str(e) for e in internal_iterations])}",
        ],
        environment_variables=environment_variables,
    )
    print(f"Submitted job with sid {statistics_collection_job}\n")
    time.sleep(1)
    print("Removing the temp.sbatch for the statistics collection file")
    pathlib.Path("temp.sbatch").unlink()


def main():
    configurations = get_configurations()
    environment_variables = get_environment_variables(configurations)
    code_directory = configurations["code_directory_path"]
    print("Starting to setup and run the experiments!")
    iterations_to_use = create_all_experiments_folders(code_directory, environment_variables, configurations)
    for experiment_index, experiment in enumerate(configurations[EXPERIMENTS_CONFIG_STR]):
        internal_iterations = iterations_to_use[experiment_index]
        experiment_sids = []
        for fold in range(configurations["num_folds"]):
            print(f"Working on fold {fold} of the experiment with domain {experiment['domain_file_name']}\n")
            for version_index, compared_version in enumerate(experiment["compared_versions"]):
                progress_bar(version_index, len(experiment["compared_versions"]))
                arguments = create_execution_arguments(experiment, fold, compared_version)
                for internal_iteration in internal_iterations:
                    arguments.append(f"--iteration_number {internal_iteration}")
                    sid = None
                    while sid is None:
                        sid = submit_job_and_validate_execution(
                            code_directory,
                            configurations,
                            experiment,
                            fold,
                            arguments,
                            environment_variables,
                            f"{experiment['domain_file_name']}_{fold}_{internal_iteration}_{learning_algorithms_map[compared_version]}_numeric_experiment_runner",
                            None,
                        )

                    experiment_sids.append(sid)
                    formatted_date_time = datetime.now().strftime("%A, %B %d, %Y %I:%M %p")
                    print(
                        f"{formatted_date_time} - submitted job with sid {sid} for algorithm {learning_algorithms_map[compared_version]} fold {fold} and iteration {internal_iteration}."
                    )
                    pathlib.Path("temp.sbatch").unlink()
                    progress_bar(version_index, len(experiment["compared_versions"]))
                    arguments.pop(-1)  # removing the internal iteration from the arguments list

            time.sleep(5)

        print("Finished building the experiment folds!")
        execute_statistics_collection_job(
            code_directory, configurations, environment_variables, experiment, experiment_sids, internal_iterations,
        )


if __name__ == "__main__":
    main()



"""Module that stores amd learns an action's numeric state fluents."""
import logging
from collections import defaultdict
from typing import Dict, List, Tuple, Optional, Set

import numpy as np
from pddl_plus_parser.models import PDDLFunction, Precondition, NumericalExpressionTree

from sam_learning.core.numeric_learning.convex_hull_learner import ConvexHullLearner
from sam_learning.core.numeric_learning.linear_regression_learner import LinearRegressionLearner
from sam_learning.core.numeric_learning.numeric_utils import create_monomials, create_grounded_monomials
from sam_learning.core.numeric_learning.incremental_convex_hull_learner import IncrementalConvexHullLearner

np.seterr(divide="ignore", invalid="ignore")


class NumericFluentStateStorage:
    """Stores and learned the numeric state fluents of a single action."""

    logger: logging.Logger
    previous_state_storage: Dict[str, List[float]]  # lifted function str -> numeric values.
    next_state_storage: Dict[str, List[float]]  # lifted function str -> numeric values.

    def __init__(
        self,
        action_name: str,
        domain_functions: Dict[str, PDDLFunction],
        polynom_degree: int = 0,
        epsilon: float = 0.0,
        qhull_options: str = "",
        incremental: bool = False,
    ):
        self.action_name = action_name
        self.logger = logging.getLogger(__name__)
        self.previous_state_storage = defaultdict(list)
        self.next_state_storage = defaultdict(list)
        self.monomials = create_monomials(list(domain_functions.keys()), polynom_degree)
        self.convex_hull_learner = (
            ConvexHullLearner(action_name, domain_functions, polynom_degree=polynom_degree, epsilon=epsilon, qhull_options=qhull_options)
            if not incremental
            else IncrementalConvexHullLearner(
                action_name, domain_functions, polynom_degree=polynom_degree, epsilon=epsilon, qhull_options=qhull_options
            )
        )

        self.linear_regression_learner = LinearRegressionLearner(action_name, domain_functions, polynom_degree=polynom_degree)

    def add_to_previous_state_storage(self, state_fluents: Dict[str, PDDLFunction]) -> None:
        """Adds the matched lifted state fluents to the previous state storage.

        :param state_fluents: the lifted state fluents that were matched for the action.
        """

        sample_dataset = create_grounded_monomials(self.monomials, state_fluents)
        self.convex_hull_learner.add_new_point(sample_dataset)
        self.linear_regression_learner.add_new_observation(sample_dataset, store_in_prev_state=True)

    def add_to_next_state_storage(self, state_fluents: Dict[str, PDDLFunction]) -> None:
        """Adds the matched lifted state fluents to the next state storage.

        :param state_fluents: the lifted state fluents that were matched for the action.
        """
        next_state_values = {state_fluent_lifted_str: state_fluent_data.value for state_fluent_lifted_str, state_fluent_data in state_fluents.items()}
        self.linear_regression_learner.add_new_observation(next_state_values, store_in_prev_state=False)

    def construct_safe_linear_inequalities(self, relevant_fluents: Optional[List[str]] = None) -> Precondition:
        """Constructs the linear inequalities strings that will be used in the learned model later.

        :param relevant_fluents: the fluents that are part of the action's preconditions and effects.
        :return: The precondition that contains the linear inequalities.
        """
        self.logger.info("Constructing the safe linear inequalities.")
        return self.convex_hull_learner.construct_safe_linear_inequalities(relevant_fluents)

    def construct_assignment_equations(
        self, allow_unsafe: bool = False, relevant_fluents: Optional[List[str]] = None
    ) -> Tuple[Set[NumericalExpressionTree], Optional[Precondition], bool]:
        """Constructs the assignment statements for the action according to the changed value functions.

        :param allow_unsafe: whether to allow unsafe fluents in the assignment equations.
        :param relevant_fluents: the fluents that are part of the action's preconditions and effects.
        :return: the constructed assignment statements.
        """
        self.logger.info("Constructing the assignment equations.")
        return self.linear_regression_learner.construct_assignment_equations(allow_unsafe=allow_unsafe, relevant_fluents=relevant_fluents)



import pathlib
import time

from experiments.cluster_scripts.common import (
    submit_job,
    get_configurations,
    get_environment_variables,
    create_internal_iterations_list,
    EXPERIMENTS_CONFIG_STR,
)


def execute_statistics_collection_job(code_directory, configuration, environment_variables, experiment, internal_iterations):
    print(f"Creating the job that will collect the statistics from all the domain's experiments.")
    statistics_collection_job = submit_job(
        conda_env="online_nsam",
        mem="4G",
        python_file=f"{code_directory}/{configuration['results_collection_script_name']}",
        jobname=f"collect_statistics_{experiment['domain_file_name']}",
        suppress_output=False,
        arguments=[
            f"--working_directory_path {experiment['working_directory_path']}",
            f"--domain_file_name {experiment['domain_file_name']}",
            f"--learning_algorithms {','.join([str(e) for e in experiment['compared_versions']])}",
            f"--num_folds {configuration['num_folds']}",
            f"--internal_iterations {','.join([str(e) for e in internal_iterations])}",
        ],
        environment_variables=environment_variables,
    )
    print(f"Submitted job with sid {statistics_collection_job}\n")
    time.sleep(1)
    print("Removing the temp.sbatch for the statistics collection file")
    pathlib.Path("temp.sbatch").unlink()


def main():
    configurations = get_configurations()
    environment_variables = get_environment_variables(configurations)
    code_directory = configurations["code_directory_path"]
    for experiment_index, experiment in enumerate(configurations[EXPERIMENTS_CONFIG_STR]):
        parallelization_data = experiment["parallelization_data"]
        internal_iterations = create_internal_iterations_list(parallelization_data)
        execute_statistics_collection_job(
            code_directory, configurations, environment_variables, experiment, internal_iterations,
        )


if __name__ == "__main__":
    main()



"""This module contains the LinearRegressionLearner class."""
import logging
from typing import Optional, List, Dict, Tuple, Set

import numpy as np
import pandas as pd
from pandas import DataFrame
from pddl_plus_parser.models import Precondition, NumericalExpressionTree, PDDLFunction
from sklearn.linear_model import LinearRegression

from sam_learning.core.exceptions import NotSafeActionError
from sam_learning.core.learning_types import EquationSolutionType, ConditionType
from sam_learning.core.numeric_learning.numeric_utils import (
    detect_linear_dependent_features,
    construct_linear_equation_string,
    construct_non_circular_assignment,
    construct_multiplication_strings,
    prettify_coefficients,
    construct_numeric_conditions,
    construct_numeric_effects,
    filter_constant_features,
    get_num_independent_equations,
    create_monomials,
    create_polynomial_string,
)

LABEL_COLUMN = "label"
NEXT_STATE_PREFIX = "next_state_"  # prefix for the next state variables.
SUBTRACTION_COLUMN = "subtraction"
POST_NEXT_STATE_PREFIX_INDEX = len(NEXT_STATE_PREFIX)

LEGAL_LEARNING_SCORE = 1.00


class LinearRegressionLearner:
    def __init__(self, action_name: str, domain_functions: Dict[str, PDDLFunction], polynom_degree: int = 0):
        self.logger = logging.getLogger(__name__)
        self.action_name = action_name
        self.domain_functions = {func.name: func for func in domain_functions.values()}
        functions = list([function.untyped_representation for function in domain_functions.values()])
        self._function_representation = functions
        monomials = create_monomials(functions, polynom_degree)
        self.previous_state_data = DataFrame(columns=[create_polynomial_string(monomial) for monomial in monomials])
        self.next_state_data = DataFrame(columns=functions)

    def _combine_states_data(self, relevant_fluents: Optional[List[str]] = None) -> DataFrame:
        """Combines the previous and next states data into a single dataframe.

        :return: the combined dataframe.
        """
        # assuming that if a function is an effect of the action it will always be present in the next state.
        if relevant_fluents is not None:
            self.previous_state_data = self.previous_state_data[
                [fluent for fluent in self.previous_state_data.columns.tolist() if fluent in relevant_fluents]
            ]
            self.next_state_data = self.next_state_data[[fluent for fluent in self.next_state_data.columns.tolist() if fluent in relevant_fluents]]

        next_state_df = self.next_state_data.add_prefix(NEXT_STATE_PREFIX)
        combined_data = pd.concat([self.previous_state_data, next_state_df], axis=1)
        # getting the columns that contain missing values in the previous state.
        combined_data.drop_duplicates(inplace=True)
        return combined_data

    def _validate_legal_equations(self, values_df: DataFrame) -> bool:
        """Validates that there are enough independent equations which enable for a single solution for the equation.

        :param values_df: the matrix constructed based on the observations.
        """
        if len(values_df) == 1:
            return False

        num_dimensions = len(values_df.columns.tolist()) + 1  # +1 for the bias.
        num_independent_rows = get_num_independent_equations(values_df)
        if num_independent_rows >= num_dimensions:
            return True

        failure_reason = (
            f"There are too few independent rows of data! " f"cannot create a linear equation from the current data for action - {self.action_name}!"
        )
        self.logger.warning(failure_reason)

        return False

    def _solve_regression_problem(
        self, values_matrix: np.ndarray, function_post_values: np.ndarray, allow_unsafe_learning: bool = True
    ) -> Tuple[List[float], float]:
        """Solves the polynomial equations using a matrix form.

        Note: the equation Ax=b is solved as: x = inverse(A)*b.

        :param values_matrix: the A matrix that contains the previous values of the function variables.
        :param function_post_values: the resulting values after the linear change.
        :param allow_unsafe_learning: whether to allow unsafe learning.
        :return: the vector representing the coefficients for the function variables and the learning score (R^2).
        """
        regressor = LinearRegression()
        regressor.fit(values_matrix, function_post_values)
        learning_score = regressor.score(values_matrix, function_post_values)
        if learning_score < LEGAL_LEARNING_SCORE:
            reason = (
                f"The learned effects are not safe since the R^2 is not high enough. "
                f"Got R^2 of {learning_score} and expected {LEGAL_LEARNING_SCORE}! "
                f"Action {self.action_name} is not safe!"
            )
            self.logger.warning(reason)
            if not allow_unsafe_learning:
                raise NotSafeActionError(self.action_name, reason, EquationSolutionType.no_solution_found)

        coefficients = list(regressor.coef_) + [regressor.intercept_]
        coefficients = prettify_coefficients(coefficients)
        self.logger.debug(f"Learned the coefficients for the numeric equations with r^2 score of {learning_score}")
        return coefficients, learning_score

    def _calculate_scale_factor(self, coefficient_vector: List[float], regression_df: DataFrame, lifted_function: str) -> Optional[float]:
        """Calculates the scale factor for the function.

        Note:
            The scale factor defines the number of times a variable increases or decreases in value.

        :param coefficient_vector: the vector of coefficients for the function variables.
        :param regression_df: the dataframe containing the regression features.
        :param lifted_function: the function to calculate the scale factor for.
        :return: the scale factor for the function.
        """
        if (
            lifted_function in regression_df.columns
            and coefficient_vector[list(regression_df.columns).index(lifted_function)] != 0
            and all(
                [
                    coefficient_vector[list(regression_df.columns).index(function)] == 0
                    for function in regression_df.columns
                    if function != lifted_function
                ]
            )
        ):
            self.logger.debug(f"The value of the {lifted_function} is being scaled by a constant factor!")
            return round(coefficient_vector[list(regression_df.columns).index(lifted_function)], 10)

        return None

    @staticmethod
    def _extract_non_zero_change(lifted_function: str, regression_df: DataFrame) -> Tuple[float, float]:
        """Extracts the non-zero change in the function.

         Note:
            The value is extracted  based on the difference between the previous and the next states.

        :param lifted_function: the name of the function to extract the change for.
        :param regression_df: the matrix containing the previous values of the function and the resulting values.
        :return: the first non-zero value of the function and the previous value.
        """
        previous_value = next_value = 0.0
        for i in range(len(regression_df)):
            previous_value = regression_df.iloc[i][lifted_function]
            next_value = regression_df.iloc[i][LABEL_COLUMN]
            if previous_value != next_value:
                break

        return next_value, previous_value

    def _solve_linear_equations(self, lifted_function: str, regression_df: DataFrame, allow_unsafe_learning: bool = False) -> Optional[str]:
        """Computes the change in the function value based on the previous values of the function.

        Note: We assume in this stage that the change results from a polynomial function of the previous values.

        :param lifted_function: the function to compute the change for.
        :param regression_df: the matrix containing the previous values of the function and the resulting values.
        :param allow_unsafe_learning: whether to allow unsafe learning.
        :return: the string representing the polynomial function change in PDDL+ format.
        """
        regression_array = np.array(regression_df.loc[:, regression_df.columns != LABEL_COLUMN])
        function_post_values = np.array(regression_df[LABEL_COLUMN])
        coefficient_vector, learning_score = self._solve_regression_problem(regression_array, function_post_values, allow_unsafe_learning)

        if all([coef == 0 for coef in coefficient_vector]) and len(regression_df[LABEL_COLUMN].unique()) == 1:
            self.logger.debug(
                "The algorithm designated a vector of zeros to the equation "
                "which means that there are no coefficients and the label is also constant value."
            )
            return f"(assign {lifted_function} {regression_df[LABEL_COLUMN].unique()[0]})"

        functions_and_dummy = list(regression_df.columns[:-1]) + ["(dummy)"]
        scale_factor = self._calculate_scale_factor(coefficient_vector, regression_df, lifted_function)
        if scale_factor is not None:
            if scale_factor == 1:
                self.logger.debug("The scale factor is 1, there is no change in the function value.")
                return None

            scale = "scale-up" if scale_factor > 1 else "scale-down"
            self.logger.debug("Currently supporting only scaling of the function by a constant value.")
            return f"({scale} {lifted_function} {scale_factor})"

        if lifted_function in regression_df.columns and coefficient_vector[list(regression_df.columns).index(lifted_function)] != 0:
            self.logger.debug("the function is a part of the equation, cannot use circular format!")
            coefficients_map = {lifted_func: coef for lifted_func, coef in zip(functions_and_dummy, coefficient_vector)}
            next_value, previous_value = self._extract_non_zero_change(lifted_function, regression_df)
            return construct_non_circular_assignment(lifted_function, coefficients_map, previous_value, next_value)

        multiplication_functions = construct_multiplication_strings(coefficient_vector, functions_and_dummy)
        if len(multiplication_functions) == 0:
            self.logger.debug(
                "The algorithm designated a vector of zeros to the equation " "which means that there are not coefficients. Continuing."
            )
            return None

        constructed_right_side = construct_linear_equation_string(multiplication_functions)
        return f"(assign {lifted_function} {constructed_right_side})"

    @staticmethod
    def _extract_const_conditions(precondition_statements: List[List[str]], additional_conditions: List[str] = None) -> List[str]:
        """Extract conditions that are constant for all the samples.

        Note:
            These preconditions that appear in all samples will be considered axioms and will be added the
            general preconditions to create a more compact version of the action's preconditions.

        :param precondition_statements: the precondition statements.
        :return: the extracted constant conditions.
        """
        const_preconditions = set(additional_conditions) if additional_conditions else set()
        for conjunction_statement in precondition_statements:
            for condition in conjunction_statement:
                if all([condition in other_conjunction for other_conjunction in precondition_statements]):
                    const_preconditions.add(condition)
                    continue

        for conjunction_statement in precondition_statements:
            for const_condition in const_preconditions:
                if const_condition in conjunction_statement:
                    conjunction_statement.remove(const_condition)

        return list(const_preconditions)

    def _construct_effect_from_single_sample(self) -> Set[NumericalExpressionTree]:
        """constructs the effect from a single sample.

        :return: the constructed numeric effect.
        """
        assignment_statements = []
        for fluent in self.next_state_data.columns.tolist():
            if (
                fluent in self.previous_state_data.columns.tolist()
                and self.previous_state_data.iloc[0][fluent] == self.next_state_data.iloc[0][fluent]
            ):
                self.logger.debug("The value of the fluent is the same before and after the application of the action.")
                continue

            if fluent not in self.previous_state_data.columns.tolist():
                self.logger.debug("Assuming that the fluent wasn't initialized and the next state is an assignment.")

            assignment_statements.append(f"(assign {fluent} {self.next_state_data.iloc[0][fluent]})")

        return construct_numeric_effects(assignment_statements, self.domain_functions)

    def add_new_observation(self, state_data: Dict[str, float], store_in_prev_state: bool = True) -> None:
        """Adds a new point to the convex hull learner.

        Note:
            This method is supposed to improve the performance of the CH calculations by incrementally adding points.

        :param state_data: the data of the state.
        :param store_in_prev_state: whether to store the data in the previous state or the next state.
        """
        state_sample = DataFrame.from_dict(data={k: [v] for k, v in state_data.items()}, orient="columns")
        data_to_update = self.previous_state_data if store_in_prev_state else self.next_state_data
        if not store_in_prev_state:
            state_sample = state_sample.drop([col for col in state_sample.columns if col not in self._function_representation], axis=1)

        attribute_name = "previous_state_data" if store_in_prev_state else "next_state_data"

        if data_to_update.empty:
            setattr(self, attribute_name, state_sample)
            return

        data_to_update = pd.concat([data_to_update, state_sample], ignore_index=True)
        data_to_update.dropna(axis="columns", inplace=True)
        setattr(self, attribute_name, data_to_update)

    def construct_assignment_equations(
        self, allow_unsafe: bool = False, relevant_fluents: Optional[List[str]] = None
    ) -> Tuple[Set[NumericalExpressionTree], Optional[Precondition], bool]:
        """Constructs the assignment statements for the action according to the changed value functions.

        :param allow_unsafe: whether to allow unsafe learning.
        :param relevant_fluents: the fluents that are part of the action's preconditions and effects.
        :return: the constructed effects with the possibly additional preconditions.
        """
        if relevant_fluents is not None and len(relevant_fluents) == 0:
            self.logger.debug(f"No numeric effects for the action {self.action_name}.")
            return set(), None, False

        if len(self.next_state_data) == 0:
            self.logger.debug(f"No observations for the action {self.action_name}.")
            return set(), None, False

        assignment_statements = []
        self.logger.info(f"Constructing the fluent assignment equations for action {self.action_name}.")
        combined_data = self._combine_states_data(relevant_fluents)
        if combined_data.shape[0] == 1:
            self.logger.info(f"The action {self.action_name} contains a single unique observation!")
            return self._construct_effect_from_single_sample(), None, False

        # The dataset contains more than one observation.
        self.logger.debug("Removing fluents that are constant...")
        tagged_next_state_fluents = [f"{NEXT_STATE_PREFIX}{fluent_name}" for fluent_name in self.next_state_data.columns.tolist()]
        filtered_df, const_features_conditions, const_features = filter_constant_features(combined_data, columns_to_ignore=tagged_next_state_fluents)

        features_df = filtered_df.copy()[[k for k in self.previous_state_data.columns.tolist() if k in filtered_df.columns]]
        regression_df, linear_dep_conditions, dependent_columns = detect_linear_dependent_features(features_df)
        combined_conditions = linear_dep_conditions + const_features_conditions

        is_safe_to_learn = self._validate_legal_equations(regression_df)
        for feature_fluent, tagged_fluent in zip(self.next_state_data.columns.tolist(), tagged_next_state_fluents):
            regression_df[LABEL_COLUMN] = combined_data[tagged_fluent]
            if self.previous_state_data[feature_fluent].equals(self.next_state_data[feature_fluent]):
                self.logger.debug(
                    f"The value of the {feature_fluent} is the same before and "
                    f"after the application of the action. The action does not change the value!"
                )
                continue

            polynomial_equation = self._solve_linear_equations(feature_fluent, regression_df, allow_unsafe_learning=allow_unsafe)

            if polynomial_equation is not None:
                assignment_statements.append(polynomial_equation)

        if not is_safe_to_learn:
            return construct_numeric_effects(assignment_statements, self.domain_functions), None, False

        return (
            construct_numeric_effects(assignment_statements, domain_functions=self.domain_functions),
            construct_numeric_conditions(combined_conditions, ConditionType.conjunctive, domain_functions=self.domain_functions,),
            True,
        )



"""Module to store types related to the learning output."""
from enum import Enum


class ConditionType(Enum):
    conjunctive = 1
    disjunctive = 2


class EquationSolutionType(Enum):
    """The solution types possible as output to the learning algorithm."""
    no_solution_found = 1
    not_enough_data = 2
    convex_hull_not_found = 3
    ok = 4



import logging
import os
import subprocess
import uuid
from pathlib import Path
from typing import Optional

VALIDATOR_DIRECTORY = Path(os.environ["VALIDATOR_DIRECTORY"])

VALID_PLAN = "Plan valid"
INAPPLICABLE_PLAN = "Plan failed to execute"
PLAN_APPLICABLE = "Plan executed successfully"
GOAL_NOT_REACHED = "Goal not satisfied"

MAX_RUNNING_TIME = 60

logger = logging.getLogger(__name__)


def run_validate_script(domain_file_path: Path, problem_file_path: Path, solution_file_path: Path) -> Path:
    """Validates that the plan for the input problem.

    :param domain_file_path: the path to the domain file.
    :param problem_file_path: the path to the problem file.
    :param solution_file_path: the path to the solution file.
    :return: the path to the validation log file.
    """
    os.chdir(VALIDATOR_DIRECTORY)
    logger.info("Running VAL to validate the plan's correctness.")
    validation_file_path = domain_file_path.parent / f"validation_log_{uuid.uuid4()}.txt"
    run_command = f"./Validate -v -t 0.1 {domain_file_path} {problem_file_path} " \
                  f"{solution_file_path} > {validation_file_path}"
    try:
        subprocess.check_output(run_command, shell=True)

    except subprocess.CalledProcessError as e:
        logger.warning(f"VAL returned status code {e.returncode}.")

    logger.info("Finished validating the solution file.")
    return validation_file_path



"""Module test for some of the matching utility functionality."""
from typing import Dict, Set

from pddl_plus_parser.lisp_parsers import DomainParser
from pddl_plus_parser.models import Domain, GroundedPredicate, State
from pytest import fixture

from sam_learning.core import extract_discrete_effects, extract_not_effects
from tests.consts import DEPOTS_NUMERIC_DOMAIN_PATH

TEST_LIFTED_SIGNATURE_ITEMS = ["?s", "?d", "?i", "?m"]
TEST_GROUNDED_ACTION_CALL = ["s1", "test_direction", "test_instrument", "test_mode"]


@fixture()
def domain() -> Domain:
    domain_parser = DomainParser(DEPOTS_NUMERIC_DOMAIN_PATH)
    return domain_parser.parse_domain()


@fixture()
def previous_state_predicates(domain: Domain) -> Dict[str, Set[GroundedPredicate]]:
    at_predicate = domain.predicates["at"]
    on_predicate = domain.predicates["on"]
    in_predicate = domain.predicates["in"]
    lifting_predicate = domain.predicates["lifting"]
    available_predicate = domain.predicates["available"]
    clear_predicate = domain.predicates["clear"]
    return {
        "(at ?x ?y)": {GroundedPredicate(name="at", signature=at_predicate.signature, object_mapping={"?x": "x1", "?y": "y1"})},
        "(on ?x ?y)": {GroundedPredicate(name="on", signature=on_predicate.signature, object_mapping={"?x": "x1", "?y": "y1"})},
        "(in ?x ?y)": {GroundedPredicate(name="in", signature=in_predicate.signature, object_mapping={"?x": "x1", "?y": "y1"})},
        "(lifting ?x ?y)": {GroundedPredicate(name="lifting", signature=lifting_predicate.signature, object_mapping={"?x": "x1", "?y": "y1"})},
        "(available ?x)": {GroundedPredicate(name="available", signature=available_predicate.signature, object_mapping={"?x": "x1"})},
        "(clear ?x)": {GroundedPredicate(name="clear", signature=clear_predicate.signature, object_mapping={"?x": "x1"})},
    }


@fixture()
def valid_previous_state(previous_state_predicates: Dict[str, Set[GroundedPredicate]]) -> State:
    return State(predicates=previous_state_predicates, fluents={})


def test_extract_effects_when_no_change_occurs_returns_empty_dictionaries(valid_previous_state: State):
    state_predicates = valid_previous_state.state_predicates
    state_predicates_set = {predicate for predicates_set in state_predicates.values() for predicate in predicates_set}
    add_effects, delete_effects = extract_discrete_effects(state_predicates_set, state_predicates_set)
    assert len(add_effects) == 0
    assert len(delete_effects) == 0


def test_extract_effects_with_only_predicate_added_adds_to_add_effects(
    valid_previous_state: State, previous_state_predicates: Dict[str, Set[GroundedPredicate]], domain: Domain
):
    clear_predicate = domain.predicates["clear"]
    next_state_predicates = {name: {*predicates} for name, predicates in previous_state_predicates.items()}
    next_state_predicates["(clear ?x)"].add(GroundedPredicate(name="clear", signature=clear_predicate.signature, object_mapping={"?x": "x2"}))
    next_state = State(predicates=next_state_predicates, fluents={})

    prev_state_predicates = valid_previous_state.state_predicates
    prev_state_predicates_set = {predicate for predicates_set in prev_state_predicates.values() for predicate in predicates_set}
    next_state_predicates = next_state.state_predicates
    next_state_predicates_set = {predicate for predicates_set in next_state_predicates.values() for predicate in predicates_set}

    add_effects, delete_effects = extract_discrete_effects(prev_state_predicates_set, next_state_predicates_set)

    assert len(add_effects) == 1
    assert len(delete_effects) == 0


def test_extract_not_effects_returns_all_predicates_of_next_state_negated(
    valid_previous_state: State, previous_state_predicates: Dict[str, Set[GroundedPredicate]], domain: Domain
):
    clear_predicate = domain.predicates["clear"]
    next_state_predicates = {name: {*predicates} for name, predicates in previous_state_predicates.items()}
    new_pred = GroundedPredicate(name="clear", signature=clear_predicate.signature, object_mapping={"?x": "x2"})
    next_state_predicates["(clear ?x)"].add(new_pred)
    next_state_predicates_set = {predicate for predicates_set in next_state_predicates.values() for predicate in predicates_set}
    not_effects = extract_not_effects(next_state_predicates_set)

    for predicate in next_state_predicates_set:
        negated_predicate = predicate.copy(is_negated=True)
        assert negated_predicate in not_effects



"""Module test for the vocabulary_creator module."""
from pddl_plus_parser.models import Domain, Problem
from pytest import fixture

from sam_learning.core import VocabularyCreator


@fixture()
def vocabulary_creator() -> VocabularyCreator:
    return VocabularyCreator()


def test_create_vocabulary_creates_grounded_predicates_only_for_those_with_matching_types(
        elevators_domain: Domain, vocabulary_creator: VocabularyCreator, elevators_problem: Problem):
    vocabulary_predicates = vocabulary_creator.create_vocabulary(
        domain=elevators_domain,
        observed_objects={"n1": elevators_problem.objects["n1"], "n2": elevators_problem.objects["n2"]})
    assert list(vocabulary_predicates.keys()) == ['(above ?floor1 ?floor2)', '(next ?n1 ?n2)']


def test_create_vocabulary_creates_grounded_predicates_when_given_two_types_of_objects(
        elevators_domain: Domain, vocabulary_creator: VocabularyCreator, elevators_problem: Problem):
    vocabulary_predicates = vocabulary_creator.create_vocabulary(
        domain=elevators_domain,
        observed_objects={"p0": elevators_problem.objects["p0"], "slow0-0": elevators_problem.objects["slow0-0"]})
    assert list(vocabulary_predicates.keys()) == ['(boarded ?person ?lift)']


def test_create_vocabulary_creates_grounded_predicates_when_constants_and_objects(
        woodworking_domain: Domain, vocabulary_creator: VocabularyCreator, woodworking_problem: Problem):
    vocabulary_predicates = vocabulary_creator.create_vocabulary(
        domain=woodworking_domain,
        observed_objects={"b0": woodworking_problem.objects["b0"],
                          "verysmooth": woodworking_domain.constants["verysmooth"],
                          "grinder0": woodworking_problem.objects["grinder0"]})
    assert set(vocabulary_predicates.keys()) == {'(surface-condition ?obj ?surface)', '(available ?obj)',
                                                 '(is-smooth ?surface)', '(has-colour ?agent ?colour)',
                                                 '(grind-treatment-change ?agent ?old ?new)'}


def test_create_lifted_vocabulary_creates_lifted_predicates_only_for_those_with_matching_types(
        woodworking_domain: Domain, vocabulary_creator: VocabularyCreator):
    vocabulary_predicates = vocabulary_creator.create_lifted_vocabulary(
        domain=woodworking_domain,
        possible_parameters=woodworking_domain.actions["do-grind"].signature)
    assert vocabulary_predicates is not None


def test_create_lifted_functions_vocabulary_creates_correct_functions(
        depot_domain: Domain, vocabulary_creator: VocabularyCreator):
    vocabulary_functions = vocabulary_creator.create_lifted_functions_vocabulary(
        domain=depot_domain,
        possible_parameters=depot_domain.actions["unload"].signature)
    assert vocabulary_functions is not None


def test_create_grounded_actions_vocabulary_creates_all_possible_assignments_of_actions(
        depot_domain: Domain, vocabulary_creator: VocabularyCreator, depot_problem: Problem):
    observed_objects = depot_problem.objects
    vocabulary_actions = vocabulary_creator.create_grounded_actions_vocabulary(
        domain=depot_domain,
        observed_objects=observed_objects)
    assert vocabulary_actions is not None
    assert len({action.name for action in vocabulary_actions}) == len(depot_domain.actions)
    print([str(action) for action in vocabulary_actions])


def test_create_grounded_actions_vocabulary_creates_only_unique_actions(
        depot_domain: Domain, vocabulary_creator: VocabularyCreator, depot_problem: Problem):
    observed_objects = depot_problem.objects
    vocabulary_actions = vocabulary_creator.create_grounded_actions_vocabulary(
        domain=depot_domain,
        observed_objects=observed_objects)
    assert len({str(action) for action in vocabulary_actions}) == len([action for action in vocabulary_actions])
    print([str(action) for action in vocabulary_actions])


def test_create_grounded_actions_vocabulary_creates_actions_that_are_applicable_in_the_state(
        minecraft_large_domain: Domain, vocabulary_creator: VocabularyCreator, minecraft_large_problem: Problem):
    observed_objects = minecraft_large_problem.objects
    vocabulary_actions = vocabulary_creator.create_grounded_actions_vocabulary(
        domain=minecraft_large_domain,
        observed_objects=observed_objects)
    action_signatures = {str(action) for action in vocabulary_actions}
    assert "(tp_to cell15 cell0)" in action_signatures



"""Module that learns polynomial preconditions and effects from a domain."""
import itertools
from typing import Dict, List, Optional

import numpy
from pddl_plus_parser.models import PDDLFunction, Precondition

from sam_learning.core.baseline_algorithms_version.naive_numeric_fluent_learner_algorithm import NaiveNumericFluentStateStorage


class NaivePolynomialFluentsLearningAlgorithm(NaiveNumericFluentStateStorage):
    """Class that learns polynomial preconditions and effects from a domain.

    Note:
        If the polynom degree is 0 the algorithm reverts to its linear version.
        degree of 1 is the multiplication of each couple of state fluents.
        degree 2 and above is the maximal degree of the polynomial.
    """
    polynom_degree: int
    is_verbose: bool

    def __init__(self, action_name: str, polynom_degree: int,
                 domain_functions: Dict[str, PDDLFunction], is_verbose: bool = False):
        super().__init__(action_name, domain_functions)
        self.polynom_degree = polynom_degree
        self.is_verbose = is_verbose

    def _create_polynomial_string_recursive(self, fluents: List[str]) -> str:
        """Creates the polynomial string representing the equation recursively.

        :param fluents: the numeric fluents to create the polynomial string from.
        :return: the polynomial string representing the equation.
        """
        if len(fluents) == 1:
            return fluents[0]

        return f"(* {fluents[0]} {self._create_polynomial_string_recursive(fluents[1:])})"

    def create_polynomial_string(self, fluents: List[str]) -> str:
        """The auxiliary function that creates the polynomial string representing the equation.

        :param fluents: the numeric fluents to create the polynomial string from.
        :return: the polynomial string representing the equation.
        """
        return self._create_polynomial_string_recursive(fluents)

    def _add_polynom_to_storage(self, state_fluents: Dict[str, PDDLFunction],
                                storage: Dict[str, List[float]]) -> None:
        """Adds the polynomial representation of the state fluents to the storage.

        :param state_fluents: the numeric fluents present in the input state.
        :param storage: the storage to update.
        """
        if self.polynom_degree == 1:
            for first_fluent, second_fluent in itertools.combinations(list(state_fluents.keys()), r=2):
                multiplied_fluent = self.create_polynomial_string(sorted([first_fluent, second_fluent]))
                storage[multiplied_fluent].append(
                    state_fluents[first_fluent].value * state_fluents[second_fluent].value)
            return

        for degree in range(2, self.polynom_degree + 1):
            for fluent_combination in itertools.combinations_with_replacement(
                    list(state_fluents.keys()), r=degree):
                polynomial_fluent = self.create_polynomial_string(sorted(list(fluent_combination)))
                values = [state_fluents[fluent].value for fluent in fluent_combination]
                self.previous_state_storage[polynomial_fluent].append(numpy.prod(values))

    def add_to_previous_state_storage(self, state_fluents: Dict[str, PDDLFunction]) -> None:
        """Adds the matched lifted state fluents to the previous state storage.

        :param state_fluents: the lifted state fluents that were matched for the action.
        """
        super().add_to_previous_state_storage(state_fluents)
        if self.polynom_degree == 0:
            return

        self._add_polynom_to_storage(state_fluents, self.previous_state_storage)

    def add_to_next_state_storage(self, state_fluents: Dict[str, PDDLFunction]) -> None:
        """Adds the matched lifted state fluents to the next state storage.

        :param state_fluents: the lifted state fluents that were matched for the action.
        """
        super().add_to_next_state_storage(state_fluents)
        if self.polynom_degree == 0:
            return

        for fluent in self.next_state_storage:
            if len(self.previous_state_storage.get(fluent, [])) != len(self.next_state_storage[fluent]):
                self.logger.debug("This is a case where effects create new fluents - should adjust the previous state.")
                self.previous_state_storage[fluent].append(0)

    def construct_safe_linear_inequalities(
            self, relevant_fluents: Optional[List[str]] = None) -> Precondition:
        """Constructs the linear inequalities strings that will be used in the learned model later.

        :return: the inequality strings and the type of equations that were constructed (injunctive / disjunctive)
        """
        algorithm_relevant_fluents = relevant_fluents or list(self.previous_state_storage.keys())
        polynomial_relevant_fluents = [*algorithm_relevant_fluents]
        if self.polynom_degree == 0:
            return super().construct_safe_linear_inequalities(relevant_fluents)

        if self.is_verbose:
            self.logger.debug(f"Fluents are given as verbose - so only learning the fluents given as input!")
            return super().construct_safe_linear_inequalities(algorithm_relevant_fluents)

        if self.polynom_degree == 1:
            for first_fluent, second_fluent in itertools.combinations(relevant_fluents, r=2):
                polynomial_relevant_fluents.append(self.create_polynomial_string([first_fluent, second_fluent]))

            return super().construct_safe_linear_inequalities(polynomial_relevant_fluents)

        for degree in range(2, self.polynom_degree + 1):
            for fluent_combination in itertools.combinations_with_replacement(relevant_fluents, r=degree):
                polynomial_relevant_fluents.append(self.create_polynomial_string(list(fluent_combination)))

        return super().construct_safe_linear_inequalities(polynomial_relevant_fluents)



"""Matches predicates to their corresponding actions based on the common types."""
import logging
from typing import List, Tuple, Optional

from pddl_plus_parser.models import Domain, Predicate, GroundedPredicate, ActionCall, PDDLObject

from sam_learning.core.matching_utils import contains_duplicates, create_signature_permutations


class PredicatesMatcher:
    """Class that matches predicates according to the needed properties in the learning process."""

    matcher_domain: Domain
    logger: logging.Logger

    def __init__(self, domain: Domain):
        self.logger = logging.getLogger(__name__)
        self.matcher_domain = domain

    @staticmethod
    def _extract_combinations_data(possible_combinations: Tuple[Tuple[str]]) -> Tuple[List[str], List[str]]:
        """extracts the data of the combinations so that the matcher would be able to use it.

        :param possible_combinations: the possible combinations of the parameters.
        :return: the extracted parameters and the objects.
        """
        call_objects = []
        lifted_params = []
        for obj, param in possible_combinations:
            call_objects.append(obj)
            lifted_params.append(param)

        return call_objects, lifted_params

    def _filter_out_impossible_combinations(
        self, grounded_predicate: GroundedPredicate, possible_matches: List[Predicate], possible_grounded_params: List[List[str]]
    ) -> List[Predicate]:
        """Filters out signature permutations that don't match the original predicate's order.

        :param grounded_predicate: the grounded predicate that is being matches.
        :param possible_matches: the possible matching permutations generated.
        :param possible_grounded_params: the list of objects that represent the lifted parameters.
            used to filter out wrong permutations.
        :return: the correct possible matches that fit the original parameters order.
        """
        self.logger.debug("Filtering out impossible matches from the matches list.")
        filtered_matches = []
        for match, grounded_matching_objects in zip(possible_matches, possible_grounded_params):
            is_possible_match = True
            if grounded_predicate.grounded_objects != grounded_matching_objects:
                continue

            for signature_item_type, grounded_predicate_signature_item_type in zip(match.signature.values(), grounded_predicate.signature.values()):
                if not grounded_predicate_signature_item_type.is_sub_type(signature_item_type):
                    is_possible_match = False
                    break

            if is_possible_match:
                filtered_matches.append(match)

        return filtered_matches

    def match_predicate_to_action_literals(
        self, grounded_predicate: GroundedPredicate, action_call: ActionCall, extra_grounded_object: str = None, extra_lifted_object: str = None
    ) -> List[Predicate]:
        """Matches the action objects to the predicate objects.

        :param grounded_predicate: the grounded predicate that was observed.
        :param action_call: the action that was called in the observation.
        :return: lifted predicates with signatures matching the action.
        :param extra_grounded_object: an additional object that is not part of the action call.
            Used in cases where we observe universal effects / preconditions.
        :param extra_lifted_object: the additional lifted object matching the grounded observed one.
        """
        self.logger.debug(
            f"Trying to match the grounded predicate - {grounded_predicate.untyped_representation} " f"to the action call {str(action_call)}"
        )
        if len(grounded_predicate.signature) == 0:
            self.logger.debug("The predicate has no parameters, by default matches the action!")
            return [Predicate(name=grounded_predicate.name, signature={}, is_positive=grounded_predicate.is_positive)]

        lifted_action_data = self.matcher_domain.actions[action_call.name]
        constants = self.matcher_domain.constants
        grounded_predicate_call = grounded_predicate.grounded_objects

        constants_in_predicate = [obj for obj in grounded_predicate_call if obj in constants]
        action_grounded_objects = (
            action_call.parameters + constants_in_predicate if extra_grounded_object is None else action_call.parameters + [extra_grounded_object]
        )
        lifted_action_params = (
            list(lifted_action_data.signature.keys()) + constants_in_predicate
            if extra_lifted_object is None
            else list(lifted_action_data.signature.keys()) + [extra_lifted_object]
        )

        if contains_duplicates(action_call.parameters):
            self.logger.debug(f"Action {str(action_call)} was executed with duplicated objects!")

        possible_parameter_permutations = create_signature_permutations(action_grounded_objects, lifted_action_params, len(grounded_predicate_call))
        possible_matches = []
        grounded_base_params = []
        for possible_permutation in possible_parameter_permutations:
            possible_match_action_objects, lifted_parameters = self._extract_combinations_data(possible_permutation)
            if set(possible_match_action_objects) == set(grounded_predicate_call):
                combined_types = {**lifted_action_data.signature}
                combined_types.update({name: constants[name].type for name in constants_in_predicate})
                if extra_lifted_object is not None:
                    self.logger.debug(f"Adding the extra lifted object {extra_lifted_object} to the signature.")
                    additional_lifted_parameter = [
                        key for key in grounded_predicate.object_mapping if grounded_predicate.object_mapping[key] == extra_grounded_object
                    ][0]
                    combined_types[extra_lifted_object] = grounded_predicate.signature[additional_lifted_parameter]

                possible_matches.append(
                    Predicate(
                        name=grounded_predicate.name,
                        signature={lifted_parameter_name: combined_types[lifted_parameter_name] for lifted_parameter_name in lifted_parameters},
                        is_positive=grounded_predicate.is_positive,
                    )
                )
                grounded_base_params.append(possible_match_action_objects)

        possible_matches = self._filter_out_impossible_combinations(grounded_predicate, possible_matches, grounded_base_params)
        return possible_matches

    def get_injective_match(self, grounded_predicate: GroundedPredicate, action_call: ActionCall) -> Optional[Predicate]:
        """Get the injective match for the predicate and the action.

        :param grounded_predicate: the grounded predicate that was observed.
        :param action_call: the action that was called in the observation.
        :return: lifted predicates with signatures matching the action.
        """
        possible_matches = self.match_predicate_to_action_literals(grounded_predicate, action_call)
        if len(possible_matches) == 1:
            lifted_match = possible_matches[0]
            lifted_match.is_positive = grounded_predicate.is_positive
            return lifted_match

        return None

    def get_possible_literal_matches(
        self,
        grounded_action_call: ActionCall,
        state_literals: List[GroundedPredicate],
        extra_grounded_object: str = None,
        extra_lifted_object: str = None,
    ) -> List[Predicate]:
        """Get a list of possible preconditions for the action according to the previous state.

        :param grounded_action_call: the grounded action that was executed according to the trajectory.
        :param state_literals: the list of literals that we try to match according to the action.
        :param extra_grounded_object: an additional object that is not part of the action call.
            Used in cases where we observe universal effects / preconditions.
        :param extra_lifted_object: the additional lifted object matching the grounded observed one.
        :return: a list of possible preconditions for the action that is being executed.
        """
        self.logger.debug(f"Finding the possible matches for the grounded action - {str(grounded_action_call)}")
        possible_matches = []
        for state_predicate in state_literals:
            if extra_grounded_object is None or extra_grounded_object not in state_predicate.grounded_objects:
                possible_matches.extend(self.match_predicate_to_action_literals(state_predicate, grounded_action_call))
                continue

            possible_matches.extend(
                self.match_predicate_to_action_literals(state_predicate, grounded_action_call, extra_grounded_object, extra_lifted_object)
            )

        return possible_matches
